# for 0.0.96.11.0.255 События связанные с напряжением СТО 34.01-5.1-006-2021 таблица Д.1
VOLTAGE_1 = "Фаза А - пропадание напряжения"
VOLTAGE_2 = "Фаза А - восстановление напряжения"
VOLTAGE_3 = "Фаза В - пропадание напряжения"
VOLTAGE_4 = "Фаза В - восстановление напряжения"
VOLTAGE_5 = "Фаза С - пропадание напряжения"
VOLTAGE_6 = "Фаза С - восстановление напряжения"
VOLTAGE_7 = "Превышение напряжения любой фазы"
VOLTAGE_8 = "Окончание перенапряжения любой фазы"
VOLTAGE_9 = "Низкое напряжение любой фазы - начало"
VOLTAGE_10 = "Низкое напряжение любой фазы - окончание"
VOLTAGE_11 = "Превышение коэффициента несимметрии напряжений по обратной последовательности - начало"
VOLTAGE_12 = "Превышение коэффициента несимметрии напряжений по обратной последовательности - окончание"
VOLTAGE_13 = "Фаза А - перенапряжение начало"
VOLTAGE_14 = "Фаза А - перенапряжение окончание"
VOLTAGE_15 = "Фаза В - перенапряжение начало"
VOLTAGE_16 = "Фаза В - перенапряжение окончание"
VOLTAGE_17 = "Фаза С - перенапряжение начало"
VOLTAGE_18 = "Фаза С - перенапряжение окончание"
VOLTAGE_19 = "Фаза А - провал начало"
VOLTAGE_20 = "Фаза А - провал окончание"
VOLTAGE_21 = "Фаза В - провал начало"
VOLTAGE_22 = "Фаза В - провал окончание"
VOLTAGE_23 = "Фаза С - провал начало"
VOLTAGE_24 = "Фаза С - провал окончание"
VOLTAGE_25 = "Неправильная последовательность фаз начало"
VOLTAGE_26 = "Неправильная последовательность фаз окончание"
VOLTAGE_27 = "Прерывание напряжения"
VOLTAGE_28 = "Восстановление напряжения"

# for current_events События связанные с током СТО 34.01-5.1-006-2021 таблица Д.2
CURRENT_1 = "Фаза А - экспорт начало"
CURRENT_2 = "Фаза А - экспорт окончание"
CURRENT_3 = "Фаза В - экспорт начало"
CURRENT_4 = "Фаза В - экспорт окончание"
CURRENT_5 = "Фаза С - экспорт начало"
CURRENT_6 = "Фаза С - экспорт окончание"
CURRENT_7 = "Обрыв трансформатора тока фазы А"
CURRENT_8 = "Восстановление трансформатора тока фазы А"
CURRENT_9 = "Обрыв трансформатора тока фазы В"
CURRENT_10 = "Восстановление трансформатора тока фазы В"
CURRENT_11 = "Обрыв трансформатора тока фазы С"
CURRENT_12 = "Восстановление трансформатора тока фазы С"
CURRENT_13 = "Разбаланс токов - начало"
CURRENT_14 = "Разбаланс токов - окончание"
CURRENT_15 = "Замыкание трансформатора тока — начало"
CURRENT_16 = "Окончание замыкания трансформатора тока"
CURRENT_17 = "Превышение тока любой фазы - начало"
CURRENT_18 = "Окончание превышения тока любой фазы"
CURRENT_19 = "Фаза А-наличие тока при отсутствии напряжения начало"
CURRENT_20 = "Фаза А-наличие тока при отсутствии напряжения окончание"
CURRENT_21 = "Фаза В - наличие тока при отсутствии напряжения начало"
CURRENT_22 = "Фаза В - наличие тока при отсутствии напряжения окончание"
CURRENT_23 = "Фаза С - наличие тока при отсутствии напряжения начало"
CURRENT_24 = "Фаза С - наличие тока при отсутствии напряжения окончание"
CURRENT_25 = "Фаза А - превышение максимального тока начало"
CURRENT_26 = "Фаза А - превышение максимального тока окончание"
CURRENT_27 = "Фаза В - превышение максимального тока начало"
CURRENT_28 = "Фаза В - превышение максимального тока окончание"
CURRENT_29 = "Фаза С - превышение максимального тока начало"
CURRENT_30 = "Фаза С - превышение максимального тока окончание"
CURRENT_31 = "Наличие тока при отсутствии напряжения обрыв нейтрали – начало"
CURRENT_32 = "Наличие тока при отсутствии напряжения обрыв нейтрали - окончание"
CURRENT_33 = "Обратный поток мощности экспорт тока в однонаправленном счётчике - начало"
CURRENT_34 = "Обратный поток мощности экспорт тока в однонаправленном счётчике - конец"
CURRENT_35 = "Разнонаправленная мощность по фазам в трёхфазном и однофазном двухэлементном счётчике - начало"
CURRENT_36 = "Разнонаправленная мощность по фазам в трёхфазном и однофазном двухэлементном счётчике - окончание"
CURRENT_37 = "Наличие тока при выключенном реле нагрузки – начало"
CURRENT_38 = "Наличие тока при выключенном реле нагрузки – окончание"

# for 0.0.96.11.2.255 События связанные с включением/выключением ПУ. коммутации реле нагрузки СТО 34.01-5.1-006-2021 таблица Д.3
COMMUTATION_1 = "Выключение питания ПУ"
COMMUTATION_2 = "Включение питания ПУ"
COMMUTATION_3 = "Выключение абонента дистанционное"
COMMUTATION_4 = "Включение абонента дистанционное"
COMMUTATION_5 = "Получение разрешения на включение абоненту"
COMMUTATION_6 = "Выключение реле нагрузки абонентом"
COMMUTATION_7 = "Включение реле нагрузки абонентом"
COMMUTATION_8 = "Выключение локальное по превышению лимита мощности"
COMMUTATION_9 = "Выключение локальное по превышению максимального тока"
COMMUTATION_10 = "Выключение локальное при воздействии магнитного поля"
COMMUTATION_11 = "Выключение локальное по превышению напряжения"
COMMUTATION_12 = "Включение локальное при возвращение напряжения в норму"
COMMUTATION_13 = "Выключение локальное по наличию тока при отсутствии напряжения"
COMMUTATION_14 = "Выключение локальное по разбалансу токов"
COMMUTATION_15 = "Выключение локальное по температуре"
COMMUTATION_16 = "Включение резервного питания"
COMMUTATION_17 = "Отключение резервного питания"
COMMUTATION_18 = "Выключение локальное при вскрытии клеммной крышки или корпуса"
COMMUTATION_19 = "Выключение реле при превышении лимитов энергии по тарифам"
COMMUTATION_20 = "Включение реле после выключения по причине превышения активной мощности"
COMMUTATION_21 = "Включение реле после выключения по причине превышения тока"
COMMUTATION_22 = "Включение реле после выключения по причине превышения небаланса токов"
COMMUTATION_23 = "Включение реле после возвращения температуры в норму"
COMMUTATION_24 = "Включение реле после возвращения магнитного поля в норму"
COMMUTATION_25 = "Выключение реле через арбитр"
COMMUTATION_26 = "Включение реле через арбитр"
COMMUTATION_27 = "Включение реле через физический блокиратор"
COMMUTATION_28 = "Выключение реле через физический блокиратор"
COMMUTATION_29 = "Полное пропадание питания ПУ"

# for 0.0.96.11.3.255 События программирования параметров ПУ СТО 34.01-5.1-006-2021 таблица Д.4
PROGRAMING_1 = "Изменение адреса или скорости обмена RS-485-1"
PROGRAMING_2 = "Изменение адреса или скорости обмена RS-485-2"
PROGRAMING_3 = "Установка времени"
PROGRAMING_4 = "Изменение параметров перехода на летнее время"
PROGRAMING_5 = "Изменение сезонного профиля тарифного расписания ТР"
PROGRAMING_6 = "Изменение недельного профиля ТР"
PROGRAMING_7 = "Изменение суточного профиля ТР"
PROGRAMING_8 = "Изменение даты активации ТР"
PROGRAMING_9 = "Активация ТР"
PROGRAMING_10 = "Изменение расчетного дня-часа РДЧ"
PROGRAMING_11 = "Изменение режима индикации параметры"
PROGRAMING_12 = "Изменение режима индикации автопереключение"
PROGRAMING_13 = "Изменение пароля низкой секретности на чтение"
PROGRAMING_14 = "Изменение пароля высокой секретности на запись"
PROGRAMING_15 = "Изменение данных точки учета"
PROGRAMING_16 = "Изменение коэффициента трансформации по току"
PROGRAMING_17 = "Изменение коэффициента трансформации по напряжению"
PROGRAMING_18 = "Изменение параметров линии для вычисления потерь в ЛЭП"
PROGRAMING_19 = "Изменение лимита мощности для отключения"
PROGRAMING_20 = "Изменение интервала времени на отключение по мощности"
PROGRAMING_21 = "Изменение интервала времени на отключение по превышению максимального тока"
PROGRAMING_22 = "Изменение интервала времени на отключение по максимальному напряжению"
PROGRAMING_23 = "Изменение интервала времени на отключение по воздействию магнитного поля"
PROGRAMING_24 = "Изменение порога для фиксации перерыва в питании"
PROGRAMING_25 = "Изменение порога для фиксации перенапряжения"
PROGRAMING_26 = "Изменение порога для фиксации провала напряжения"
PROGRAMING_27 = "Изменение порога для фиксации превышения тангенса"
PROGRAMING_28 = "Изменение порога для фиксации коэффициента несимметрии напряжений"
PROGRAMING_29 = "Изменение согласованного напряжения"
PROGRAMING_30 = "Изменение интервала интегрирования пиковой мощности"
PROGRAMING_31 = "Изменение периода захвата профиля PROGRAMING_1 ="
PROGRAMING_32 = "Изменение периода захвата профиля 2"
PROGRAMING_33 = "Изменение режима подсветки LCD"
PROGRAMING_34 = "Изменение режима телеметрии 1"
PROGRAMING_35 = "Очистка месячного журнала"
PROGRAMING_36 = "Очистка суточного журнала"
PROGRAMING_37 = "Очистка журнала напряжения"
PROGRAMING_38 = "Очистка журнала тока"
PROGRAMING_39 = "Очистка журнала вкл/выкл"
PROGRAMING_40 = "Очистка журнала внешних воздействий"
PROGRAMING_41 = "Очистка журнала соединений"
PROGRAMING_42 = "Очистка журнала несанкционированного доступа"
PROGRAMING_43 = "Очистка журнала качества сети"
PROGRAMING_44 = "Очистка журнала тангенса"
PROGRAMING_45 = "Очистка журнала входов/выходов"
PROGRAMING_46 = "Очистка профиля 1"
PROGRAMING_47 = "Очистка профиля 2"
PROGRAMING_48 = "Очистка профиля 3"
PROGRAMING_49 = "Изменение таблицы специальных дней"
PROGRAMING_50 = "Изменение режима управления реле"
PROGRAMING_51 = "Фиксация показаний в месячном журнале"
PROGRAMING_52 = "Изменение режима инициативного выхода"
PROGRAMING_53 = "Изменение одноадресного ключа для низкой секретности"
PROGRAMING_54 = "Изменение широковещательного ключа шифрования для низкой секретности"
PROGRAMING_55 = "Изменение одноадресного ключа для высокой секретности"
PROGRAMING_56 = "Изменение широковещательного ключа для высокой секретности"
PROGRAMING_57 = "Изменение ключа аутентификации для высокой секретности"
PROGRAMING_58 = "Изменение мэстер-ключа"
PROGRAMING_59 = "Изменение уровня преобразования для низкой секретности"
PROGRAMING_60 = "Изменение уровня преобразования для высокой секретности"
PROGRAMING_61 = "Изменение номера дистанционного дисплея"
PROGRAMING_62 = "Изменение режима учета активной энергии по модулю или раздельно в двух направлениях"
PROGRAMING_63 = "Установка времени по GPS/ГЛОНАСС"
PROGRAMING_64 = "Изменение режима отключения по обрыву нейтрали"
PROGRAMING_65 = "Обновление ПО"
PROGRAMING_66 = "Изменение режима отключения по разбалансу токов"
PROGRAMING_67 = "Изменение режима отключения по температуре"
PROGRAMING_68 = "Коррекция времени"
PROGRAMING_69 = "Изменение ключа аутентификации для низкой секретности"
PROGRAMING_70 = "Очистка флагов инициативного выхода"
PROGRAMING_71 = "Изменение таймаута для HDLC-соединения"
PROGRAMING_72 = "Изменение часов болъших нагрузок"
PROGRAMING_73 = "Изменение часов контроля максимума"
PROGRAMING_74 = "Изменение схемы подключения"
PROGRAMING_75 = "Изменение режима телеметрии 2"
PROGRAMING_76 = "Изменение режима телеметрии 3"
PROGRAMING_77 = "Изменение режима телеметрии 4"
PROGRAMING_78 = "Изменение режима отключения при вскрытии клеммной крышки или корпуса"
PROGRAMING_79 = "Изменение настройки активного коммуникационного профиля для портов связи"
PROGRAMING_80 = "Очистка журнала качества сети на месячном интервале"
PROGRAMING_81 = "Изменение интервала интегрирования параметров сети"
PROGRAMING_82 = "Изменение порогового значения по времени. Коэффициент реактивной мощности tg <р средний по всем фазам"
PROGRAMING_83 = "Изменение порогового значения по времени. Дифференциальньм ток. % от величины наибольшего ток"
PROGRAMING_84 = "Изменение порогового значения по времени. Коэффициент несимметрии по обратной последовательности"
PROGRAMING_85 = "Изменение адреса или скорости обмена Оптопорт Р1"
PROGRAMING_86 = "Изменение адреса или скорости обмена Порт Р4"
PROGRAMING_87 = "Изменение фильтра событий отключения реле нагрузки"
PROGRAMING_88 = "Изменение монитора событий реле нагрузки"
PROGRAMING_89 = "Изменение настройки учёта энергии"
PROGRAMING_90 = "Изменение порогового значения отклонения частоты"
PROGRAMING_91 = "Изменение порогового значения контроля активной мощности на интервале интегрирования"
PROGRAMING_92 = "Изменение порогового значения контроля активной мощности на интервале интегрирования в часы пиковых нагрузок"
PROGRAMING_93 = "Изменение времени фиксации стоп кадра. Фиксация стоп кадра"
PROGRAMING_94 = "Монитор событий реле нагрузки"
PROGRAMING_95 = "Монитор событий реле сигнализации 1"
PROGRAMING_96 = "Монитор событий реле сигнализации 2"
PROGRAMING_97 = "Монитор событий реле сигнализации 3"
PROGRAMING_98 = "Монитор событий реле сигнализации 4"
PROGRAMING_99 = "Изменение параметров арбитра реле нагрузки"
PROGRAMING_100 = "Изменение параметров арбитра реле сигнализации 1"
PROGRAMING_101 = "Изменение параметров арбитра реле сигнализации 2"
PROGRAMING_102 = "Изменение параметров арбитра реле сигнализации 3"
PROGRAMING_103 = "Изменение параметров арбитра реле сигнализации 4"
PROGRAMING_104 = "Изменение фильтра событий реле сигнализации 1"
PROGRAMING_105 = "Изменение фильтра событий реле сигнализации 2"
PROGRAMING_106 = "Изменение фильтра событий реле сигнализации 3 "
PROGRAMING_107 = "Изменение фильтра событий реле сигнализации 4"
PROGRAMING_108 = "Изменение режима управления реле сигнализации 1"
PROGRAMING_109 = "Изменение режима управления реле сигнализации 2"
PROGRAMING_110 = "Изменение режима управления реле сигнализации 3"
PROGRAMING_111 = "Изменение режима управления реле сигнализации 4"
PROGRAMING_112 = "Изменение типа контакта реле сигнализации"
PROGRAMING_113 = "Изменение таймаута для TCP/UDP соединения (Оптопорт P1)"
PROGRAMING_114 = "Изменение таймаута для TCP/UDP соединения (Порт P2)"
PROGRAMING_115 = "Изменение таймаута для TCP/UDP соединения (Порт P3)"
PROGRAMING_116 = "Изменение таймаута для TCP/UDP соединения (Порт P4)"
PROGRAMING_117 = "Очистка журнала «выхода тангенса за порог на интервале интегрирования»"
PROGRAMING_118 = "Очистка журнала «коррекции времени»"
PROGRAMING_119 = "Очистка журнала «На начало года»"
PROGRAMING_120 = "Резерв"
PROGRAMING_121 = "Очистка журнала «Контроля мощности»"
PROGRAMING_122 = "Очистка журнала «Батареи»"
PROGRAMING_123 = "Очистка журнала «Контроль блокиратора реле нагрузки»"
PROGRAMING_124 = "Очистка журнала «Контроль температуры»"
PROGRAMING_125 = "Очистка журнала «Отклонение напряжения фазы А»"
PROGRAMING_126 = "Очистка журнала «Отклонение напряжения фазы B»"
PROGRAMING_127 = "Очистка журнала «Отклонение напряжения фазы C»"
PROGRAMING_128 = "Очистка журнала «Отклонение линейного напряжения AB»"
PROGRAMING_129 = "Очистка журнала «Отклонение линейного напряжения BC»"
PROGRAMING_130 = "Очистка журнала «Отклонение линейного напряжения CA»"
PROGRAMING_131 = "Очистка журнала «Превышение напряжения»"
PROGRAMING_132 = "Очистка журнала «Прерывание напряжения»"
PROGRAMING_133 = "Очистка журнала «Телесигнализация»"
PROGRAMING_134 = "Очистка журнала «Нештатная ситуация сети»"
PROGRAMING_135 = "Изменение порога напряжения по нулевой последовательности, время до срабатывания события, с (нештатная ситуация сети)"
PROGRAMING_136 = "Изменение порога напряжения по нулевой последовательности, время задержки установки события PUSH, с (нештатная ситуация сети)"
PROGRAMING_137 = "Изменение порога напряжения, минимальное значение, В(нештатная ситуация сети)"
PROGRAMING_138 = "Изменение порога напряжения, время до срабатывания события, с(нештатная ситуация сети)"
PROGRAMING_139 = "Изменение порога напряжения, время задержки установки события PUSH, с (нештатная ситуация сети)"
PROGRAMING_140 = "Изменение порога напряжения по обратной последовательности, максимальное значение, В (нештатная ситуация сети)"
PROGRAMING_141 = "Изменение порога напряжения по обратной последовательности, время до срабатывания события, с (нештатная ситуация сети)"
PROGRAMING_142 = "Изменение порога напряжения по обратной последовательности, время задержки установки события PUSH, с (нештатная ситуация сети)"
PROGRAMING_143 = "Обжатие электронных пломб"
PROGRAMING_144 = "Очистка фиксации событий воздействия магнитного и/или ВЧ поля"
PROGRAMING_145 = "Изменение часового пояса"
PROGRAMING_146 = "Изменение последовательности вывода на ЖКИ в режиме «Автопрокрутка»"
PROGRAMING_147 = "Изменение последовательности вывода на ЖКИ в режиме «По кнопке»"
PROGRAMING_148 = "Изменение уровня лимита по току"
PROGRAMING_149 = "Изменение уровня лимита по напряжению"
PROGRAMING_150 = "Номер аварийного тарифа"
PROGRAMING_1024 = "КПЗ. Изменение конфигурации ПУ"

# for 0.0.96.11.4.255 События внешних воздействий СТО 34.01-5.1-006-2021 таблица Д.5
EXTERNAL_IMPACT_1 = "Магнитное поле - начало"
EXTERNAL_IMPACT_2 = "Магнитное поле - окончание"
EXTERNAL_IMPACT_3 = "Срабатывание электронной пломбы крышки клеммников"
EXTERNAL_IMPACT_4 = "Срабатывание электронной пломбы корпуса"
EXTERNAL_IMPACT_5 = "Срабатывание электронной пломбы внешнего датчика"
EXTERNAL_IMPACT_6 = "Воздействие ВЧ поля - начало"
EXTERNAL_IMPACT_7 = "Воздействие ВЧ поля- окончание"

# for 0.0.96.11.5.255 Коммуникационные события СТО 34.01-5.1-006-2021 таблица Д.6
COMMUNICATION_1 = "Разорвано соединение интерфейс"
COMMUNICATION_2 = "Установлено соединение интерфейс"

# for 0.0.96.11.6.255 События контроля доступа СТО 34.01-5.1-006-2021 таблица Д.7
ACCESS_1 = "Попытка несанкционированного доступа интерфейс"
ACCESS_2 = "Нарушение требований протокола"
ACCESS_3 = "Блокировка по превышению количества неправильных паролей"
ACCESS_4 = "Ошибка верификации прошивки"

# for 0.0.96.11.7.255 Коды событий для журнала самодиагностики СТО 34.01-5.1-006-2021 таблица Д.8
SELF_DIAGNOSTIC_1 = "Инициализация ПУ"
SELF_DIAGNOSTIC_2 = "Измерительный блок — ошибка"
SELF_DIAGNOSTIC_3 = "Измерительный блок — норма"
SELF_DIAGNOSTIC_4 = "Вычислительный блок — ошибка"
SELF_DIAGNOSTIC_5 = "Часы реального времени — ошибка"
SELF_DIAGNOSTIC_6 = "Часы реального времени — норма"
SELF_DIAGNOSTIC_7 = "Блок питания — ошибка"
SELF_DIAGNOSTIC_8 = "Блок питания — норма"
SELF_DIAGNOSTIC_9 = "Дисплей — ошибка"
SELF_DIAGNOSTIC_10 = "Дисплей — норма"
SELF_DIAGNOSTIC_11 = "Блок памяти — ошибка"
SELF_DIAGNOSTIC_12 = "Блок памяти — норма"
SELF_DIAGNOSTIC_13 = "Блок памяти программ — ошибка"
SELF_DIAGNOSTIC_14 = "Блок памяти программ — норма"
SELF_DIAGNOSTIC_15 = "Система тактирования ядра — ошибка"
SELF_DIAGNOSTIC_16 = "Система тактирования ядра — норма"
SELF_DIAGNOSTIC_17 = "Система тактирования часов — ошибка"
SELF_DIAGNOSTIC_18 = "Система тактирования часов — норма"
SELF_DIAGNOSTIC_19 = "Вычислительный блок — норма"

# for 0.0.96.11.8.255 События по превышению реактивной мощности tg ф тангенс сети СТО 34.01-5.1-006-2021 таблица Д.9
REACTIVE_POWER_1 = "Превышение установленного порога — начало"
REACTIVE_POWER_2 = "Превышение установленного порога — окончание"

# for 0.0.96.5.1.255 Статус качества сети профиль суточных показаний
POWER_QUALITY1_1 = "Отклонение напряжения более чем на 10% от номинала"
POWER_QUALITY1_2 = "Отклонение частоты более чем на 04 Гц от номинала"

# for POWER_QUALITY1_0 =.0.96.5.4.255 Статус качества сети журнал качества сети
POWER_QUALITY2_1 = "Снижение напряжения более чем на 10%"
POWER_QUALITY2_8 = "Повышение напряжения более чем на 10%"
POWER_QUALITY2_0x10 = "Снижение частоты более чем на 04 Гц"
POWER_QUALITY2_0x20 = "Снижение частоты более чем на 02 Гц"
POWER_QUALITY2_0x40 = "Увеличение частоты более чем на 02 Гц"
POWER_QUALITY2_0x80 = "Увеличение частоты более чем на 04 Гц"
POWER_QUALITY2_0x4000 = "Снижение частоты более, чем на заданный порог"
POWER_QUALITY2_0x8000 = "Увеличение частоты более, чем на заданный порог"



from ...types import common_data_types as cdt
from ...cosem_interface_classes import cosem_interface_class as ic, overview


class ClassId(cdt.LongUnsigned):
    """ Class ID type """

    def validate(self):
        if not int(self) in overview.ClassID.get_all_id():
            raise ValueError(F'Unknown DLMS class with ID {int(self)}')



from .__class_init__ import *


class Status(cdt.Enum):
    """ Indicates the registration status of the  modem. """
    ELEMENTS = {b'\x00': en.NOT_REGISTERED,
                b'\x01': en.REGISTERED_HOME_NETWORK,
                b'\x02': en.NOT_REGISTERED_BUT_MT,
                b'\x03': en.REGISTRATION_DENIED,
                b'\x04': en.UNKNOWN,
                b'\x05': en.REGISTERED_ROAMING}


class CSAttachment(cdt.Enum):
    """ Indicates the current circuit switched status."""
    ELEMENTS = {b'\x00': en.INACTIVE,
                b'\x01': en.INCOMING_CALL,
                b'\x02': en.ACTIVE}


class PSStatus(cdt.Enum):
    """ Indicates the packet switched status of the modem. """
    ELEMENTS = {b'\x00': en.INACTIVE,
                b'\x01': en.GPRS,
                b'\x02': en.EDGE,
                b'\x03': en.UMTS,
                b'\x04': en.HSDPA}


class SignalQuality(cdt.Unsigned):
    """for string report"""
    @property
    def report(self) -> str:
        value = int(self)
        if value == 0:
            return F"–113 dBm {en.OR_LESS}(0)"
        elif value == 1:
            return F"–111 dBm(1)"
        elif value < 31:
            return F"{-109+(value-2)*2} dBm({value})"
        elif value == 31:
            return F"–51 dBm {en.OR_GREATER}(31)"
        elif value == 99:
            return F"{en.NOT_KNOWN_OR_NOT_DETECTABLE}"
        else:
            return F"wrong {value=}"


class CellInfoType(cdt.Structure):
    """ Params of element """
    values: tuple[cdt.LongUnsigned, cdt.LongUnsigned, SignalQuality, cdt.Unsigned]
    ELEMENTS = (cdt.StructElement(cdt.se.CELL_ID, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.LOCATION_ID, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.SIGNAL_QUALITY, SignalQuality),
                cdt.StructElement(cdt.se.BER, cdt.Unsigned))

    @property
    def cell_ID(self) -> cdt.LongUnsigned:
        """Two-byte cell ID in hexadecimal format"""
        return self.values[0]

    @property
    def location_ID(self) -> cdt.LongUnsigned:
        """Two-byte location area code (LAC) in hexadecimal format"""
        return self.values[1]

    @property
    def signal_quality(self) -> SignalQuality:
        """represents the signal quality"""
        return self.values[2]

    @property
    def ber(self) -> cdt.Unsigned:
        """Bit error (BER) measurement in percent:
        (0...7) as RXQUAL_n values specified in ETSI GSM 05.08 8.2.4
        (99) not known or not detectable."""
        return self.values[3]


class AdjacentCellInfo(cdt.Structure):
    values: tuple[cdt.LongUnsigned, cdt.Unsigned]
    ELEMENTS = (cdt.StructElement(cdt.se.CELL_ID, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.SIGNAL_QUALITY, cdt.Unsigned))

    @property
    def cell_ID(self) -> cdt.LongUnsigned:
        """Two-byte cell ID in hexadecimal format"""
        return self.values[0]

    @property
    def signal_quality(self) -> cdt.Unsigned:
        """represents the signal quality:
            0: -113 dBm or less,
            1: -11q dBm,
            2..30: -109...53 dBm,
            31: -51 or greater,
            99 not known or not detectable"""
        return self.values[1]


class AdjacentCells(cdt.Array):
    TYPE = AdjacentCellInfo


class GSMDiagnostic(ic.COSEMInterfaceClasses):
    """ The GSM/GPRS network is undergoing constant changes in terms of registration status, signal quality etc. It is necessary to monitor and log the relevant parameters in order
     to obtain diagnostic information that allows identifying communication problems in the network. An instance of the 'GSM diagnostic' class stores parameters of the GSM/GPRS
     network necessary for analysing the operation of the network."""
    NAME = cn.GSM_DIAGNOSTIC
    CLASS_ID = ClassID.GSM_DIAGNOSTIC
    VERSION = Version.V0
    A_ELEMENTS = (ic.ICAElement(an.OPERATOR, cdt.VisibleString),
                  ic.ICAElement(an.STATUS, Status, 0, 255, 0),
                  ic.ICAElement(an.CS_ATTACHMENT, CSAttachment, 0, 255, 0),
                  ic.ICAElement(an.PS_STATUS, PSStatus, 0, 255, 0),
                  ic.ICAElement(an.CELL_INFO, CellInfoType),
                  ic.ICAElement(an.ADJACENT_CELL, AdjacentCells),
                  ic.ICAElement(an.CAPTURE_TIME, cdt.DateTime))

    def characteristics_init(self):
        """nothing do it"""

    @property
    def operator(self) -> cdt.VisibleString:
        return self.get_attr(2)

    @property
    def status(self) -> Status:
        return self.get_attr(3)

    @property
    def cs_attachment(self) -> CSAttachment:
        return self.get_attr(4)

    @property
    def ps_status(self) -> PSStatus:
        return self.get_attr(5)

    @property
    def cell_info(self) -> CellInfoType:
        return self.get_attr(6)

    @property
    def adjacent_cell(self) -> AdjacentCells:
        return self.get_attr(7)

    @property
    def capture_time(self) -> cdt.DateTime:
        return self.get_attr(8)





OPTICAL_PORT = "Optical port"
RS485 = "RS-485"
TCP_IP = "TCP/IP"
NOT_DEFINED = "not defined"
INTERNAL_CRYSTAL = "internal crystal"
MAINS_FREQUENCY_50_HZ = "mains frequency 50 Hz"
MAINS_FREQUENCY_60_HZ = "mains frequency 60 Hz"
GPS_GLOBAL_POSITIONING_SYSTEM = "GPS (global positioning system)"
RADIO_CONTROLLED = "radio controlled"
CONTROL_MODE_0 = "None. The disconnect control object is always in <connected> state"
CONTROL_MODE_1 = "Disconnection: remote(b, c), manual(f), local(g) Reconnection: remote(d), manual(e)"
CONTROL_MODE_2 = "Disconnection: remote(b, c), manual(f), local(g) Reconnection: remote(a), manual(e)"
CONTROL_MODE_3 = "Disconnection: remote(b, c), local(g) Reconnection: Remote(d), manual(e)"
CONTROL_MODE_4 = "Disconnection: remote(b, c), local(g) Reconnection: Remote(a), manual(e)"
CONTROL_MODE_5 = "Disconnection: remote(b, c), manual(f), local(g) Reconnection: remote(d), manual(e), local(h)"
CONTROL_MODE_6 = "Disconnection: remote(b, c), local(g) Reconnection: remote(d), manual(e), local(h)"
RU_CONTROL_MODE_129 = "Disconnection: remote(b, c), manual(f), local(m) Reconnection: remote(d), manual(e)"
RU_CONTROL_MODE_130 = "Disconnection: remote(b, c), manual(f), local(m) Reconnection: remote(a), manual(e)"
RU_CONTROL_MODE_131 = "Disconnection: remote(b, c), local(m) Reconnection: remote(d), manual(e)"
RU_CONTROL_MODE_132 = "Disconnection: remote(b, c), local(m) Reconnection: remote(a)"
RU_CONTROL_MODE_133 = "Disconnection: remote(b, c), local(m) Reconnection: remote(a), local(k)"
RU_CONTROL_MODE_134 = "Disconnection: remote(b, c), local(m, s) Reconnection: remote(d), manual(e), local(p)"
DISCONNECTED = "Disconnected"
CONNECTED = "Connected"
READY_FOR_RECONNECTION = "Ready for reconnection"
NOT_REGISTERED = "not registered"
REGISTERED_HOME_NETWORK = "registered, home network"
NOT_REGISTERED_BUT_MT = "not registered, but MT is currently searching a new operator to register to"
REGISTRATION_DENIED = "registration denied"
UNKNOWN = "unknown"
REGISTERED_ROAMING = "registered, roaming"
INACTIVE = "inactive"
INCOMING_CALL = "incoming call"
ACTIVE = "active"
GPRS = "GPRS"
EDGE = "EDGE"
UMTS = "UMTS"
HSDPA = "HSDPA"
BAUD_300 = "300 baud"
BAUD_600 = "600 baud"
BAUD_1200 = "1200 baud"
BAUD_2400 = "2400 baud"
BAUD_4800 = "4800 baud"
BAUD_9600 = "9600 baud"
BAUD_19200 = "19200 baud"
BAUD_38400 = "38400 baud"
BAUD_57600 = "57600 baud"
BAUD_115200 = "115200 baud"
IMAGE_TRANSFER_NOT_INITIATED = "Image transfer not initiated"
IMAGE_TRANSFER_INITIATED = "Image transfer initiated"
IMAGE_VERIFICATION_INITIATED = "Image verification initiated"
IMAGE_VERIFICATION_SUCCESSFUL = "Image verification successful"
IMAGE_VERIFICATION_FAILED = "Image verification failed"
IMAGE_ACTIVATION_INITIATED = "Image activation initiated"
IMAGE_ACTIVATION_SUCCESSFUL = "Image activation successful"
IMAGE_ACTIVATION_FAILED = "Image activation failed"
FIFO = "fifo (first in first out)"
LIFO = "lifo (last in first out)"
LARGEST = "largest"
SMALLEST = "smallest"
NEAREST_TO_ZERO = "nearest to zero"
FAREST_FROM_ZERO = "farest from zero"
WRITE_ATTRIBUTE = "write attribute"
EXECUTE_SPECIFIC_METHOD = "execute specific method"
SCHEDULE_TYPE_1 = "size of execution time = 1; wildcard in date allowed"
SCHEDULE_TYPE_2 = "size of execution time = n; all time values are the same, wildcard in date not allowed"
SCHEDULE_TYPE_3 = "size of execution time = n; all time values are the same, wildcard in date are allowed"
SCHEDULE_TYPE_4 = "size of execution time = n; time values may be different, wildcard in date not allowed"
SCHEDULE_TYPE_5 = "size of execution time = n; time values may be different, wildcard in date are allowed"
NO_ACCESS = "no access"
READ_ONLY = "read only"
WRITE_ONLY = "write only"
READ_AND_WRITE = "read and write"
NON_ASSOCIATED = "non associated"
ASSOCIATION_PENDING = "association pending"
ASSOCIATED = "associated"
NO_STATION = "no station"
CLIENT_MANAGEMENT_PROCESS = "Client Management Process"
PUBLIC_CLIENT = "Public client"
METER_READER = "Meter Reader"
UTILITY_SETTING = "Utility Settings"
PUSH = "Push"
FIRMWARE_UPDATE = "Firmware Update"
IHD = "IHD"
NONE = "NONE"
LOW = "LOW"
HIGH = "HIGH"
HIGH_MD5 = "HIGH MD5"
HIGH_SHA1 = "HIGH SHA1"
HIGH_GMAC = "HIGH GMAC"
HIGH_SHA256 = "HIGH SHA256"
HIGH_ECDSA = "HIGH ECDSA"
ACCESS = "access"
AUTHENTICATED_ACCESS = "authenticated_access"
NOTHING = "nothing"
ALL_MESSAGES_TO_BE_AUTHENTICATION = "all messages to be authentication"
ALL_MESSAGES_TO_BE_ENCRYPTED = "all messages to be encrypted"
ALL_MESSAGES_TO_BE_AUTH_AND_ENC = "all messages to be authentication and encrypted"
RESERVED = "reserved"
AUTHENTICATION_REQUEST = "authentication request"
ENCRYPTION = "encryption"
AUTHENTICATION_AND_ENCRYPTION = "authentication and encryption"
DIGITAL_SIGNATURE = "digital signature"
ENCRYPTED_REQUEST = "encrypted request"
DIGITALLY_SIGNED_REQUEST = "digitally signed request"
AUTHENTICATION_RESPONSE = "authentication response"
ENCRYPTED_RESPONSE = "encrypted response"
DIGITALLY_SIGNED_RESPONSE = "digitally signed response"
UNUSED = "unused, shall be set 0"
AUTHENTICATION = "authentication"
AUTHENTICATION_READ_ONLY = "authenticated_read_only"
AUTHENTICATION_WRITE_ONLY = "authenticated_write_only"
AUTHENTICATION_READ_AND_WRITE = "authenticated_read_and_write"
TIME_YEAR = "a"
TIME_MONTH = "mo"
TIME_WEEK = "wk"
TIME_DAY = "d"
TIME_HOUR = "h"
TIME_MINUTE = "min"
TIME_SECOND = "s"
PHASE_ANGLE_DEGREE = "°"
TEMPERATURE = "°C"
LOCAL_CURRENCY = "currency"
LENGTH_METRE = "m"
SPEED_METRE_PER_SECOND = "m/s"
VOLUME_VALUE_CUBIC_METRE = "m³"
CORRECTED_VOLUME_CUBIC_METRE = "m³*"
VOLUME_FLUX_CUBIC_METRE_PER_HOUR = "m³/h"
CORRECTED_VOLUME_FLUX_CUBIC_METRE_PER_HOUR = "m³/h*"
VOLUME_FLUX_CUBIC_METRE_PER_DAY = "m³/d"
CORRECTED_VOLUME_FLUX_CUBIC_METRE_PER_DAY = "m³/d"
VOLUME_LITRE = "l"
MASS_KILOGRAM = "kg"
FORCE_NEWTON = "N"
ENERGY_NEWTON_METER = "Nm"
PRESSURE_PASCAL = "Pa"
PRESSURE_BAR = "bar"
ENERGY_JOULE = "J"
THERMAL_POWER_RATE_OF_CHANGE_JOULE_PER_HOUR = "J/h"
ACTIVE_POWER_WATT = "W"
APPARENT_POWER_VOLT_AMPERE = "VA"
REACTIVE_POWER_VAR = "var"
ACTIVE_ENERGY_VALUE_WATT_HOUR = "Wh"
APPARENT_ENERGY_VALUE_VOLT_AMPERE_HOUR = "VAh"
REACTIVE_ENERGY_VALUE_VAR_HOUR = "varh"
CURRENT_AMPERE = "A"
ELECTRICAL_CHARGE_COULOMB = "C"
VOLTAGE_VOLT = "V"
ELECTRIC_FIELD_STRENGTH_VOLT_PER_METRE = "V/m"
CAPACITANCE_FARAD = "F"
RESISTANCE_OHM = "Ω"
RESISTIVITY = "Ωm²/m"
MAGNETIC_FLUX_WEBER = "Wb"
MAGNETIC_FLUX_DENSITY_TESLA = "T"
MAGNETIC_FIELD_STRENGTH_AMPERE_PER_METRE = "A/m"
INDUCTANCE_HENRY = "H"
FREQUENCY_HERTZ = "Hz"
RW_ACTIVE_ENERGY_VALUE = "1/(Wh)"
RB_REACTIVE_ENERGY_VALUE = "1/(varh)"
RS_APPARENT_ENERGY_VALUE = "1/(VAh)"
VOLT_SQUARED_HOUR_VOLT_SQUARED_HOUR_METER = "V²h"
AMPERE_SQUARED_HOUR = "A²h"
MASS_FLUX_KILOGRAM_PER_SECOND = "kg/s"
CONDUCTANCE_SIEMENS = "S, mho"
TEMPERATURE_KELVIN = "K"
RU2H_VOLT_SQUARED_HOUR_METER = "1/(V²h)"
RI2H_AMPERE_SQUARED_HOUR_METER = "1/(A²h)"
RV_METER = "1/m³"
PERCENTAGE = "%"
AMPERE_HOURS = "Ah"
ENERGY_PER_VOLUME = "Wh/m³"
CALORIFIC_VALUE_WOBBE = "J/m³"
MOLAR_FRACTION_OF_GAS_COMPOSITION_MOLE_PERCENT = "Mol %"
MASS_DENSITY_QUANTITY_OF_MATERIAL = "g/m³"
DYNAMIC_VISCOSITY_PASCAL_SECOND = "Pa s"
SPECIFIC_ENERGY_JOULE_KILOGRAM = "J/kg"
PRESSURE_GRAM_PER_SQUARE_CENTIMETER = "g/cm²"
PRESSURE_ATMOSPHERE = "atm"
SIGNAL_STRENGTH_MILLIWATT = "dBm"
SIGNAL_STRENGTH_MICROVOLT = "dbµV"
LOGARITHMIC_UNIT = "dB"
LENGTH_INCH = "in"
LENGTH_FOOT = "ft"
MASS_POUND = "lb"
TEMPERATURE_FAHRENHEIT = "°F"
TEMPERATURE_RANKINE = "°R"
AREA_SQUARE_INCH = "sq. in"
AREA_SQUARE_FOOT = "sq ft"
AREA_ACRE = "ac"
VOLUME_CUBIC_INCH = "cu in"
VOLUME_CUBIC_FOOT = "cu ft"
VOLUME_ACRE_FOOT = "ac ft"
VOLUME_GALLON_IMPERIAL = "gal(imp)"
VOLUME_GALLON_US = "gal(US)"
FORCE_POUND_FORCE = "lbf"
PRESSURE_POUND_FORCE_PER_SQUARE_INCH = "psi"
DENSITY_POUND_PER_CUBIC_FOOT = "lb/cu ft"
DYNAMIC_VISCOSITY = "lb/(ft .s)"
KINEMATIC_VISCOSITY_SQUARE_FOOT_PER_SECOND = "sq ft/s"
ENERGY_BRITISH_THERMAL_UNIT = "Btu"
ENERGY_THERM_EU = "thm(EC)"
ENERGY_THERM_US = "thm(US)"
CALORIFIC_VALUE_OF_MASS_LENTHALPY = "Btu/lb"
CALORIFIC_VALUE_OF_VOLUME_WOBBE = "Btu/cu ft"
VOLUME_METER_CUBIC_FEET_RV = "cu ft rv"
SPEED_FOOT_PER_SECOND = "ft/s"
VOLUME_FLUX_CUBIC_FOOT_PER_SECOND = "cu ft/s"
VOLUME_FLUX_CUBIC_FOOT_PER_MIN = "cu ft/min"
VOLUME_FLUX_CUBIC_FOOT_PER_HOUR = "cu ft/h"
VOLUME_FLUX_CUBIC_FOOT_PER_DAY = "cu ft/d"
VOLUME_FLUX_ACRE_FOOT_PER_SECOND = "ac ft/s"
VOLUME_FLUX_ACRE_FOOT_PER_MIN = "ac ft/min"
VOLUME_FLUX_ACRE_FOOT_PER_HOUR = "ac ft/h"
VOLUME_FLUX_ACRE_FOOT_PER_DAY = "ac ft/d"
VOLUME_IMPERIAL_GALLON_METER_RV = "gal(imp) rv"
VOLUME_FLUX_IMPERIAL_GALLON_PER_SECOND = "gal(imp)/s"
VOLUME_FLUX_IMPERIAL_GALLON_PER_MIN = "gal(imp)/min"
VOLUME_FLUX_IMPERIAL_GALLON_PER_HOUR = "gal(imp)/h"
VOLUME_FLUX_IMPERIAL_GALLON_PER_DAY = "gal(imp)/d"
VOLUME_US_GALLON_RV = "gal(US) rv"
VOLUME_FLUX_US_GALLON_PER_SECOND = "gal(US)/s"
VOLUME_FLUX_US_GALLON_PER_MIN = "gal(US)/min"
VOLUME_FLUX_US_GALLON_PER_HOUR = "gal(US)/h"
VOLUME_FLUX_US_GALLON_PER_DAY = "gal(US)/d"
ENERGY_FLOW_BRITISH_THERMAL_UNIT_PER_SECOND = "Btu/s"
ENERGY_FLOW_BRITISH_THERMAL_UNIT_PER_MINUTE = "Btu/min"
ENERGY_FLOW_BRITISH_THERMAL_UNIT_PER_HOUR = "Btu/h"
ENERGY_FLOW_BRITISH_THERMAL_UNIT_PER_DAY = "Btu/d"
EXTENDED_TABLE_OF_UNITS = "extended"
OTHER_UNIT = "other"
NO_UNIT_UNITLESS_COUNT = "count"
GUEK = "global unicast encryption key"
GBEK = "global broadcast encryption key"
GAK = "authentication key"
KEK = "master key"
AES_GCM_128 = "AES-GCM-128 for authenticated encryption and AES-128 for key wrapping"
AES_GCM_128_ECDSA_P_256 = "AES-GCM-128 authenticated encryption, ECDSA P-256 digital signature, ECDH P-256 key agreement, SHA-256 hash, V.44 compression and AES-128 key wrap"
AES_GCM_256_ECDSA_P_384 = "AES-GCM-256 authenticated encryption, ECDSA P-384 digital signature, ECDH P-384 key agreement, SHA-384 hash, V.44 compression and AES-256 key wrap"
SERVER = "server"
CLIENT_OR_THIRD_PARTY = "client or third party"
CERTIFICATION_AUTHORITY = "certification authority"
OTHER = "other"
KEY_AGREEMENT = "key agreement"
TLS = "TLS"
DIGITAL_SIGNATURE_KEY_PAIR = "digital signature key pair"
KEY_AGREEMENT_KEY_PAIR = "key agreement key pair"
TLS_KEY_PAIR = "tls key pair"
CERTIFICATE_IDENTIFICATION_ENTITY = "certificate identification entity"
CERTIFICATE_IDENTIFICATION_SERIAL = "certificate identification serial"
RESTRICTION_BY_DATE = "restriction by date"
RESTRICTION_BY_ENTRY = "restriction by entry"
TCP = "TCP"
UDP = "UDP"
FTP_RESERVED = "reserved for FTP"
SMTP_RESERVED = "reserved for SMTP"
SMS = "SMS"
HDLC = "HDLC"
M_BUS = "M_BUS"
ZIGBEE_RESERVED = "reserved for ZIGBEE"
DLMS_GATEWAY = "DLMS Gateway"
MANUFACTURER_SPECIFIC = "Manufacturer specific"
A_XDR = "A-XDR encoded xDLMS APDU"
XML = "XML encoded xDLMS APDU"
UNCONFIRMED_FAILURE = "unconfirmed, retry on supporting protocol layer failure"
UNCONFIRMED_MISSING = "unconfirmed, retry on missing supporting protocol layer confirmation"
CONFIRMED = "confirmed, retry on missing confirmation"
MASTER_KEY = "master key"
GLOB_UNI_ENCR_KEY = "global unicast encryption key"
GLOB_BROAD_ENCR_KEY = "global broadcast encryption key"
IDENTIFIED_KEY = "identified key"
WRAPPED_KEY = "wrapped key"
AGREED_KEY = "agreed key"

# for GSM_diagnostic.cell_info
OR_LESS = "or less"
OR_GREATER = "or greater"
NOT_KNOWN_OR_NOT_DETECTABLE = "not known or not detectable"



from enum import Enum, IntEnum, auto
from typing import Self
from .settings import get_current_language, Language

match get_current_language():
    case Language.ENGLISH: from .Values.EN import enum_names as en
    case Language.RUSSIAN: from .Values.RU import enum_names as en


class ITEEnum(Enum):
    def is_ok(self) -> bool:
        match self:
            case self.OK: return True
            case _:       return False

    @property
    def value(self):
        return super(ITEEnum, self).value[1]

    @property
    def importance(self):
        return super(ITEEnum, self).value[0]


class Transmit(ITEEnum):
    """ error enumerator for transmit . Priority for status has biggest value """
    OK = 0, 'Успех'
    NO_PORT = 2, 'Отсутствует порт'
    TIMEOUT = 3, 'Таймаут'
    NO_ACCESS = 4, 'Нет доступа'
    NO_TRANSPORT = 5, 'Отсутствует транспортный интерфейс'
    WRITE_ERROR = 6, 'Ошибка записи'
    READ_ERROR = 7, 'Ошибка чтения'
    EXECUTE_ERROR = 8, 'Ошибка исполнения'
    ABORT = 9, 'Прерывание'
    """ Manual interrupt """
    UNKNOWN = 10, 'Неизвестная ошибка'


class Application(ITEEnum):
    """ error enumerator for application . Priority for status has biggest value """
    OK = 0, 'Успех'
    ID_ERROR = 2, 'Ошибка серийного номера'
    """ serial number identification error """
    VERSION_ERROR = 2, 'Ошибка версии'
    """ Any errors associated with version """
    MISSING_OBJ = 2, 'Отсутствует объект'
    """ requested object was missing in collection """
    EMPTY_OBJ = 2, 'Пустые данные'
    """ empty attribute in object """
    NO_CONFIG = 2, 'Конфигурация не найдена'
    VALUE_ERROR = 2, 'Ошибка значения'
    TYPE_ERROR = 2, 'Ошибка типа прибора'
    VERIFY_ERROR = 3, 'Ошибка верификации'
    ACTIVATION_ERROR = 3, 'Ошибка активации'
    RESULT_ERROR = 3, "Ошибка данных"  # COSEMpdu_GB83.asn


class TagsName(Enum):
    ROOT = 'Configure'
    LANGUAGE = 'Language'
    GROUP = 'group'
    DEVICE = 'device'
    NAME = 'name'
    ID = 'id'
    SERIAL_NUMBER = 'serial_number'
    SERIAL_NUMBER_AMOUNT = 'serial_number_amount'
    MECHANISM_ID = 'mechanism_id'
    INTERFACE = 'Interface'
    CHILD_ID = 'child_id'
    DEVICE_ROOT = 'Objects'
    TEMPLATE_ROOT = 'template.objects'


class DBOperation(IntEnum):
    WRITE_ATTR = 1
    WRITE_METH = auto()
    CREATE_DEV_TYPE = auto()


class Interface(IntEnum):
    NO_DEFINED = 0
    OPTO = 1
    RS485 = 2
    PLC = 3
    GSM = 4
    NBIO = 5
    ETHERNET = 6
    RF = 7
    LORA = 8
    WIFI = 9
    BLUETOOTH = 10
    ZIGBEE = 11
    RESERVED_12 = 12
    RESERVED_13 = 13
    RESERVED_14 = 14
    RESERVED_15 = 15
    RESERVED_16 = 16
    RESERVED_17 = 17
    RESERVED_18 = 18
    RESERVED_19 = 19
    RESERVED_20 = 20
    RESERVED_21 = 21
    RESERVED_22 = 22
    RESERVED_23 = 23
    RESERVED_24 = 24
    RESERVED_25 = 25
    RESERVED_26 = 26
    RESERVED_27 = 27
    RESERVED_28 = 28
    RESERVED_29 = 29
    MANUFACTURE_30 = 30
    MANUFACTURE_31 = 31

    @staticmethod
    def from_str(value: str) -> Self:
        return next(filter(lambda i: i.name == value, Interface))


class ChannelNumber(IntEnum):
    RESERVED = 0
    OPTO_P1 = 1
    P2 = 2
    P3 = 3
    P4 = 4
    OTHER = 5
    INNER = 6
    SERVER_KEY = 7

    @staticmethod
    def from_str(value: str) -> Self:
        return next(filter(lambda i: i.name == value, ChannelNumber))


class ConnectionType(IntEnum):
    NO_STATION = 0,
    CLIENT_MANAGEMENT_PROCESS = 1
    PUBLIC_CLIENT = 0x10
    METER_READER = 0x20
    UTILITY_SETTING = 0x30
    PUSH = 0x40
    FIRMWARE_UPDATE = 0x50
    IHD = 0x60

    def __str__(self):
        match self:
            case self.NO_STATION:                return en.NO_STATION
            case self.CLIENT_MANAGEMENT_PROCESS: return en.CLIENT_MANAGEMENT_PROCESS
            case self.PUBLIC_CLIENT:             return en.PUBLIC_CLIENT
            case self.METER_READER:              return en.METER_READER
            case self.UTILITY_SETTING:           return en.UTILITY_SETTING
            case self.PUSH:                      return en.PUSH
            case self.FIRMWARE_UPDATE:           return en.FIRMWARE_UPDATE
            case self.IHD:                       return en.IHD
            case _:                              return str(self)

    @classmethod
    def from_str(cls, value: str) -> Self:
        match value:
            case en.NO_STATION:                return cls.NO_STATION
            case en.CLIENT_MANAGEMENT_PROCESS: return cls.CLIENT_MANAGEMENT_PROCESS
            case en.PUBLIC_CLIENT:             return cls.PUBLIC_CLIENT
            case en.METER_READER:              return cls.METER_READER
            case en.UTILITY_SETTING:           return cls.UTILITY_SETTING
            case en.PUSH:                      return cls.PUSH
            case en.FIRMWARE_UPDATE:           return cls.FIRMWARE_UPDATE
            case en.IHD:                       return cls.IHD
            case _:                            return cls(int(value))


class MechanismId(IntEnum):
    NONE = 0
    LOW = 1
    HIGH = 2
    HIGH_MD5 = 3
    HIGH_SHA1 = 4
    HIGH_GMAC = 5
    HIGH_SHA256 = 6
    HIGH_ECDSA = 7

    @classmethod
    def from_str(cls, value: str) -> Self:
        return next(filter(lambda i: i.name == value, cls))


class ContextId(IntEnum):
    LN_NO_CIPHERING = 1
    SN_NO_CIPHERING = 2
    LN_CIPHERING = 3
    SN_CIPHERING = 4

    @classmethod
    def from_str(cls, value: str) -> Self:
        return next(filter(lambda i: i.name == value, cls))


if __name__ == '__main__':
    a = Transmit.OK
    print(a.name, a.value, a.importance)
    print(a.is_ok())
    match a:
        case Transmit.OK: print(a.value)
        case Application.ID_ERROR: print('id')





"""all realisation cdt.Integer subclasses"""
from ...types import common_data_types as cdt


class Only0(cdt.Integer):
    """ Limited Integer only 0 """
    NAME = F'{cdt.tn.INTEGER}(0)'

    def validate(self):
        if self.decode() != 0:
            raise ValueError(F'The integer(0) must only 0,  got {self.decode()}')



from ...types import common_data_types as cdt
from ...settings import get_current_language, Language

match get_current_language():
    case Language.ENGLISH: from ...Values.EN import enum_names as en
    case Language.RUSSIAN: from ...Values.RU import enum_names as en


class CommSpeed(cdt.Enum):
    """ The communication speed supported by the corresponding port. This communication speed can be overridden if the HDLC mode of a devive is entered through a special mode
    of another protocol. """
    ELEMENTS = {b'\x00': en.BAUD_300,
                b'\x01': en.BAUD_600,
                b'\x02': en.BAUD_1200,
                b'\x03': en.BAUD_2400,
                b'\x04': en.BAUD_4800,
                b'\x05': en.BAUD_9600,
                b'\x06': en.BAUD_19200,
                b'\x07': en.BAUD_38400,
                b'\x08': en.BAUD_57600,
                b'\x09': en.BAUD_115200}

    def decode(self) -> int:
        """ override enum key to enum value"""
        match self.contents:
            case b'\x00':  return 300
            case b'\x01':  return 600
            case b'\x02':  return 1200
            case b'\x03':  return 2400
            case b'\x04':  return 4800
            case b'\x05':  return 9600
            case b'\x06':  return 19200
            case b'\x07':  return 38400
            case b'\x08':  return 57600
            case b'\x09':  return 115200
            case _ as err: raise ValueError(F'Wrong CommSpeed: {err}, expected {CommSpeed.ELEMENTS.keys()}')

    def from_int(self, value: int) -> bytes:
        """ additional cases with Speed Values"""
        match value:
            case 300:    return b'\x00'
            case 600:    return b'\x01'
            case 1200:   return b'\x02'
            case 2400:   return b'\x03'
            case 4800:   return b'\x04'
            case 9600:   return b'\x05'
            case 19200:  return b'\x06'
            case 38400:  return b'\x07'
            case 57600:  return b'\x08'
            case 115200: return b'\x09'
            case other:  return super(CommSpeed, self).from_int(other)


class RestrictionType(cdt.Enum):
    ELEMENTS = {b'\x00': en.NONE,
                b'\x01': en.RESTRICTION_BY_DATE,
                b'\x02': en.RESTRICTION_BY_ENTRY}


class KeyInfoType(cdt.Enum):
    ELEMENTS = {b'\x00': en.IDENTIFIED_KEY,
                b'\x01': en.WRAPPED_KEY,
                b'\x02': en.AGREED_KEY}


class ProtectionType(cdt.Enum):
    ELEMENTS = {b'\x00': en.AUTHENTICATION,
                b'\x01': en.ENCRYPTED_REQUEST,
                b'\x02': en.WRAPPED_KEY,
                b'\x03': en.AGREED_KEY}



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection
from src.DLMS_SPODES.version import AppVersion


class TestType(unittest.TestCase):

    def test_dummy_class(self):
        a = collection.ProfileGenericVer0("0.0.1.0.0.255")
        print(a)

    def test_get_type_from_class(self):
        ln = cst.LogicalName("0.0.1.0.0.255")
        value = collection.get_interface_class(ut.CosemClassId(7), cdt.Unsigned(0))
        v1 = value(ln)
        print(value, v1)

    def test_get_instance(self):
        col = collection.Collection()
        col.manufacturer = b"KPZ"
        col.server_ver = AppVersion(1, 4, 0)
        col.set_spec()
        ver_obj = col.add(class_id=ut.CosemClassId(1),
                          version=cdt.Unsigned(0),
                          logical_name=cst.LogicalName("0.0.96.1.6.255"))
        ver_obj.set_attr(2, "33 30")
        inst = col.get_instance(class_id=ut.CosemClassId(3),
                                version=cdt.Unsigned(0),
                                ln=cst.LogicalName("1.0.131.35.0.255"))
        self.assertRaises(ValueError, col.get_instance, class_id=ut.CosemClassId(1),
                          version=cdt.Unsigned(0),
                          ln=cst.LogicalName("1.0.1.7.0.255"))
        inst = col.get_instance(class_id=ut.CosemClassId(1),
                                version=cdt.Unsigned(0),
                                ln=cst.LogicalName("0.128.96.13.1.255"))
        print(inst)

    def test_ClassMap(self):
        print(hash(collection.DataMap))

    def test_xmlCodingDecoding(self):
        col = collection.Collection()
        col.from_xml("09054d324d5f33_0906312e342e3130.typ")
        print(col)
        col.to_xml("test.xml")
        col.save_type("test.typ")
        ver = col.get_object("0.0.0.2.1.255")
        col.clear_server_ver()
        ver.value.set("312e312e33")
        print(ver)



# for 0.0.96.11.0.255 События связанные с напряжением ГОСТ Р 58940—2020 таблица Д.1
VOLTAGE_1 = "Phase L1-voltage drop"
VOLTAGE_2 = "Phase L1-voltage recovery"
VOLTAGE_3 = "Phase L2-voltage drop"
VOLTAGE_4 = "Phase L2-voltage recovery"
VOLTAGE_5 = "Phase L3-voltage drop"
VOLTAGE_6 = "Phase L3-voltage recovery"
VOLTAGE_7 = "Exceeding the voltage of any phase"
VOLTAGE_8 = "End of over voltage of any phase"
VOLTAGE_9 = "Low voltage of any phase-start"
VOLTAGE_10 = "Low voltage of any phase-end"
VOLTAGE_11 = "Excess of the voltage unbalance coefficient in the reverse sequence-the beginning"
VOLTAGE_12 = "Excess of the voltage unbalance coefficient in the reverse sequence - end"
VOLTAGE_13 = "Phase L1-over voltage start"
VOLTAGE_14 = "Phase L1-over voltage end"
VOLTAGE_15 = "Phase L2-over voltage start"
VOLTAGE_16 = "Phase L2-over voltage end"
VOLTAGE_17 = "Phase L3-over voltage start"
VOLTAGE_18 = "Phase L3-over voltage end"
VOLTAGE_19 = "Phase L1-failure start"
VOLTAGE_20 = "Phase L1-failure ending"
VOLTAGE_21 = "Phase L2-failure start"
VOLTAGE_22 = "Phase L2-failure ending"
VOLTAGE_23 = "Phase L3-failure start"
VOLTAGE_24 = "Phase L3-failure ending"
VOLTAGE_25 = "Incorrect phase sequence start"
VOLTAGE_26 = "Incorrect phase sequence ending"
VOLTAGE_27 = "Voltage interruption"
VOLTAGE_28 = "Voltage recovery"

# for current_events События связанные с током ГОСТ Р 58940—2020 таблица Д.2
CURRENT_1 = "Phase L1-export start"
CURRENT_2 = "Phase L1-Export end"
CURRENT_3 = "Phase L2-export start"
CURRENT_4 = "Phase L2-Export end"
CURRENT_5 = "Phase L3-export start"
CURRENT_6 = "Phase L3-Export end"
CURRENT_7 = "Phase L1 current transformer breakage"
CURRENT_8 = "Restoring the Phase L1 current transformer"
CURRENT_9 = "Phase L2 current transformer breakage"
CURRENT_10 = "Restoring the Phase L2 current transformer"
CURRENT_11 = "Phase L3 current transformer breakage"
CURRENT_12 = "Restoring the Phase L3 current transformer"
CURRENT_13 = "Current unbalance-the beginning"
CURRENT_14 = "Current unbalance-end"
CURRENT_15 = "Current transformer short circuit-start"
CURRENT_16 = "Current transformer short circuit termination"
CURRENT_17 = "Exceeding the current of any phase-the beginning"
CURRENT_18 = "End of excess current of any phase"
CURRENT_19 = "Phase L1-the presence of current in the absence of voltage start"
CURRENT_20 = "Phase L1-the presence of current in the absence of voltage end"
CURRENT_21 = "Phase L2-the presence of current in the absence of voltage start"
CURRENT_22 = "Phase L2-the presence of current in the absence of voltage end"
CURRENT_23 = "Phase L3-the presence of current in the absence of voltage start"
CURRENT_24 = "Phase L3-the presence of current in the absence of voltage end"
CURRENT_25 = "Phase L1-exceeding the maximum current start"
CURRENT_26 = "Phase L1-exceeding the maximum current end"
CURRENT_27 = "Phase L2-exceeding the maximum current start"
CURRENT_28 = "Phase L2-exceeding the maximum current end"
CURRENT_29 = "Phase L3-exceeding the maximum current start"
CURRENT_30 = "Phase L3-exceeding the maximum current end"
CURRENT_31 = "The presence of current in the absence of voltage neutral breakage. Begin"
CURRENT_32 = "The presence of current in the absence of voltage neutral breakage. End"
CURRENT_33 = "Reverse of the power export of current in a unidirectional counter. Begin"
CURRENT_34 = "Reverse of the power export of current in a unidirectional counter. End"
CURRENT_35 = "Unspressed phase power in a three -phase and single -phase two -element counter. Begin"
CURRENT_36 = "Unspressed phase power in a three -phase and single -phase two -element counter. End"
CURRENT_37 = "The presence of current with the load relay is turned off. Begin"
CURRENT_38 = "The presence of current with the load relay is turned off. End"


# for 0.0.96.11.2.255 События связанные с включением/выключением ПУ. коммутации реле нагрузки ГОСТ Р 58940—2020 таблица Д.3
COMMUTATION_1 = "Power down of the device"
COMMUTATION_2 = "Powering up the device"
COMMUTATION_3 = "Remote subscriber shutdown"
COMMUTATION_4 = "Enabling the subscriber remotely"
COMMUTATION_5 = "Getting permission to enable the subscriber"
COMMUTATION_6 = "Switching off the load relay by the subscriber"
COMMUTATION_7 = "Switching on the load relay by the subscriber"
COMMUTATION_8 = "Local shutdown when the power limit is exceeded"
COMMUTATION_9 = "Local shutdown when the maximum current is exceeded"
COMMUTATION_10 = "Local shutdown when exposed to a magnetic field"
COMMUTATION_11 = "Local shutdown when the voltage is exceeded"
COMMUTATION_12 = "Switching on local when the voltage returns to normal"
COMMUTATION_13 = "Local shutdown by the presence of current in the absence of voltage"
COMMUTATION_14 = "Local shutdown based on current unbalance"
COMMUTATION_15 = "Local temperature shutdown"
COMMUTATION_16 = "Enabling backup power"
COMMUTATION_17 = "Switching off the backup power supply"
COMMUTATION_18 = "Local shutdown when opening the terminal cover or housing"
COMMUTATION_19 = "Turning off the relay when the energy limits are exceeded by tariffs"
COMMUTATION_20 = "Turning on the relay after turning off due to the excess of active power"
COMMUTATION_21 = "Turning on the relay after turning off due to current excess"
COMMUTATION_22 = "Turning on the relay after turning off due to excess of the non -balance of currents"
COMMUTATION_23 = "Turning the relay after returning the temperature to normal"
COMMUTATION_24 = "The relay is turned on after the magnetic field returned to normal"
COMMUTATION_25 = "Turning off the relay through the arbiter"
COMMUTATION_26 = "Turning the relay through the arbiter"
COMMUTATION_27 = "Turning the relay through a physical lock"
COMMUTATION_28 = "Turning off the relay through a physical lock"
COMMUTATION_29 = "Full loss of power supply"

# for 0.0.96.11.3.255 События программирования параметров ПУ ГОСТ Р 58940—2020 таблица Д.4
PROGRAMING_1 = "Changing the address or the RS-485-1 exchange rate"
PROGRAMING_2 = "Changing the address or the RS-485-2 exchange rate"
PROGRAMING_3 = "Setting the time"
PROGRAMING_4 = "Changing Daylight Saving time settings"
PROGRAMING_5 = "Changing the seasonal profile of the tariff schedule"
PROGRAMING_6 = "Changing the weekly profile of the tariff schedule"
PROGRAMING_7 = "Changing the day profile of the tariff schedule"
PROGRAMING_8 = "Changing the activation date of the tariff schedule"
PROGRAMING_9 = "Activation of the tariff schedule"
PROGRAMING_10 = "Changing the billing day-hour"
PROGRAMING_11 = "Changing the display mode parameters"
PROGRAMING_12 = "Changing the display mode auto-switch"
PROGRAMING_13 = "Changing a low-security read-only password"
PROGRAMING_14 = "Changing the high-security password on the record"
PROGRAMING_15 = "Changing accounting point data"
PROGRAMING_16 = "Change in the current conversion factor"
PROGRAMING_17 = "Change in the voltage transformation coefficient"
PROGRAMING_18 = "Changing the line parameters for calculating power line losses"
PROGRAMING_19 = "Changing the power limit for disconnection"
PROGRAMING_20 = "Changing the power-off time interval"
PROGRAMING_21 = "Changing the time interval for switching off when the maximum current is exceeded"
PROGRAMING_22 = "Changing the time interval for switching off at the maximum voltage"
PROGRAMING_23 = "Changing the time interval for switching off due to the influence of a magnetic field"
PROGRAMING_24 = "Changing the threshold for fixing a power break"
PROGRAMING_25 = "Changing the threshold for fixing the overvoltage"
PROGRAMING_26 = "Changing the threshold for fixing the voltage drop"
PROGRAMING_27 = "Changing the threshold for fixing the excess of the tangent"
PROGRAMING_28 = "Changing the threshold for fixing the stress asymmetry coefficient"
PROGRAMING_29 = "Changing the matched voltage"
PROGRAMING_30 = "Changing the peak power integration interval"
PROGRAMING_31 = "Changing the capture period of profile PROGRAMING_1 ="
PROGRAMING_32 = "Changing the capture period of profile 2"
PROGRAMING_33 = "Changing the LCD backlight mode"
PROGRAMING_34 = "Changing the telemetry mode 1"
PROGRAMING_35 = "Cleaning the monthly log"
PROGRAMING_36 = "Clearing the daily log"
PROGRAMING_37 = "Clearing the voltage log"
PROGRAMING_38 = "Clearing the current log"
PROGRAMING_39 = "Clearing the on/off log"
PROGRAMING_40 = "Clearing the external impact log"
PROGRAMING_41 = "Clearing the connection log"
PROGRAMING_42 = "Clearing the unauthorized access log"
PROGRAMING_43 = "Clearing the network quality log"
PROGRAMING_44 = "Clearing the tangent log"
PROGRAMING_45 = "Clearing the I/O log"
PROGRAMING_46 = "Clearing profile 1"
PROGRAMING_47 = "Clearing profile 2"
PROGRAMING_48 = "Clearing profile 3"
PROGRAMING_49 = "Changing the special Days table"
PROGRAMING_50 = "Changing the relay control mode"
PROGRAMING_51 = "Recording of readings in the monthly log"
PROGRAMING_52 = "Changing the initiative exit mode"
PROGRAMING_53 = "Changing the unicast key for low security"
PROGRAMING_54 = "Changing the broadcast encryption key for low security"
PROGRAMING_55 = "Changing the unicast key for high security"
PROGRAMING_56 = "Changing the broadcast key for high security"
PROGRAMING_57 = "Changing the authentication key for high security"
PROGRAMING_58 = "Changing the master key"
PROGRAMING_59 = "Change the conversion level for low secrecy"
PROGRAMING_60 = "Change the conversion level for high security"
PROGRAMING_61 = "Changing the remote display number"
PROGRAMING_62 = "Changing the active energy metering mode modulo or separately in two directions"
PROGRAMING_63 = "Setting the time by GPS/GLONASS"
PROGRAMING_64 = "Changing the neutral break-off mode"
PROGRAMING_65 = "Software Update"
PROGRAMING_66 = "Changing the current unbalance shutdown mode"
PROGRAMING_67 = "Changing the temperature shutdown mode"
PROGRAMING_68 = "Time correction"
PROGRAMING_69 = "Changing the authentication key for low security"
PROGRAMING_70 = "Clearing Initiative exit flags"
PROGRAMING_71 = "Changing the timeout for an HDLC connection"
PROGRAMING_72 = "Changing the hours of heavy loads"
PROGRAMING_73 = "Changing the maximum monitoring hours"
PROGRAMING_74 = "Changing the connection scheme"
PROGRAMING_75 = "Changing the telemetry mode 2"
PROGRAMING_76 = "Changing the telemetry mode 3"
PROGRAMING_77 = "Changing the telemetry mode 4"
PROGRAMING_78 = "Changing the shutdown mode when opening the terminal cover or housing"
PROGRAMING_79 = "Changing the active communication profile setting for communication ports"
PROGRAMING_80 = "Clearing the network quality log on a monthly interval"
PROGRAMING_81 = "Changing the interval for integrating network parameters"
PROGRAMING_82 = "Changing the threshold value over time. Reactive power factor tg <p average for all phases"
PROGRAMING_83 = "Changing the threshold value over time. Differential current. % of the maximum current value"
PROGRAMING_84 = "Changing the threshold value over time. The coefficient of asymmetry in the reverse sequence"
PROGRAMING_85 = "Changing the address or exchange rate Optical port P1"
PROGRAMING_86 = "Changing the address or exchange rate Port P4"
PROGRAMING_87 = "Changing the filter of events of shutdown of the load relay"
PROGRAMING_88 = "Changing the monitor of events of the load relay"
PROGRAMING_89 = "Changing the setting of energy accounting"
PROGRAMING_90 = "Changing the threshold value of frequency deviations"
PROGRAMING_91 = "Changing the threshold value of active power control at the integration interval"
PROGRAMING_92 = "Changing the threshold value of active power control at the integration interval during peak loads"
PROGRAMING_93 = "Changing the time of fixing the foot frame. Fixation of the foot frame"
PROGRAMING_94 = "Monitor of the events of the load relay"
PROGRAMING_95 = "Alarm relay monitor 1"
PROGRAMING_96 = "Alarm relay monitor 2"
PROGRAMING_97 = "Alarm relay monitor 3"
PROGRAMING_98 = "Alarm relay monitor 4"
PROGRAMING_99 = "Changing the parameters of the arbiter of the load relay"
PROGRAMING_100 = "Changing the parameters of the arbiter of the alarm relay 1"
PROGRAMING_101 = "Changing the parameters of the arbiter of the alarm relay 2"
PROGRAMING_102 = "Changing the parameters of the arbiter of the alarm relay 3"
PROGRAMING_103 = "Changing the parameters of the arbiter of the alarm relay 4"
PROGRAMING_104 = "Changing the filter of events of the signaling relay 1"
PROGRAMING_105 = "Changing the filter of events of the signaling relay 2"
PROGRAMING_106 = "Changing the filter of events of the signaling relay 3"
PROGRAMING_107 = "Changing the filter of events of the signaling relay 4"
PROGRAMING_108 = "Changing the control mode of the alarm relay 1"
PROGRAMING_109 = "Changing the control mode of the alarm relay 2"
PROGRAMING_110 = "Changing the control mode of the alarm relay 3"
PROGRAMING_111 = "Changing the control mode of the alarm relay 4"
PROGRAMING_112 = "Change type of contact of the signaling relayChange type of contact of the signaling relay"
PROGRAMING_113 = "Changing the timeout for TCP/UDP connection (optical port P1)"
PROGRAMING_114 = "Changing the timeout for TCP/UDP connection (port P1)"
PROGRAMING_115 = "Changing the timeout for TCP/UDP connection (port P1)"
PROGRAMING_116 = "Changing the timeout for TCP/UDP connection (port P1)"
PROGRAMING_117 = "Cleaning the log 'Tangens output for the threshold at the integration interval'"
PROGRAMING_118 = "Cleaning the log 'Correction of Time'"
PROGRAMING_119 = "Cleaning the log 'At the beginning of the year'"
PROGRAMING_120 = "reserved"
PROGRAMING_121 = "Cleaning the log 'Power Control'"
PROGRAMING_122 = "Cleaning the log 'Battery'"
PROGRAMING_123 = "Cleaning the log 'Control of the Load Relay Relay'"
PROGRAMING_124 = "Cleaning the log 'Temperature control'"
PROGRAMING_125 = "Cleaning the log 'Deviation of phase A voltage'"
PROGRAMING_126 = "Cleaning the log 'Deviation of phase B voltage'"
PROGRAMING_127 = "Cleaning the log 'Deviation of phase C voltage'"
PROGRAMING_128 = "Cleaning the log 'Deviation of linear voltage AB'"
PROGRAMING_129 = "Cleaning the log 'Deviation of linear voltage BC'"
PROGRAMING_130 = "Cleaning the log 'Deviation of linear voltage CA'"
PROGRAMING_131 = "Cleaning the log 'Voltage Excess'"
PROGRAMING_132 = "Cleaning the log 'Interruption of voltage'"
PROGRAMING_133 = "Cleaning the log 'Television signaling'"
PROGRAMING_134 = "Cleaning the log 'Network situation'"
PROGRAMING_135 = "Changing the voltage threshold by zero sequence, time until the event is triggered, sec (Networks abnormal)"
PROGRAMING_136 = "Changing the voltage threshold through the zero sequence, the time of delaying the installation of the Push event, sec (the abnormal network situation)"
PROGRAMING_137 = "Changing the voltage threshold, minimum value, V (an emergency situations)"
PROGRAMING_138 = "Changing the voltage threshold, time before the event is triggered, sec (the abnormal network situation)"
PROGRAMING_139 = "Changing the voltage threshold, time delaying the installation of the Push event, sec (the abnormal network situation)"
PROGRAMING_140 = "Changing the voltage threshold by reverse sequence, maximum value, V (the abnormal network situation)"
PROGRAMING_141 = "Changing the voltage threshold by reverse sequence, time until the event is triggered, sec (an emergency system of the network)"
PROGRAMING_142 = "Changing the voltage threshold by the reverse sequence, the time of delaying the installation of the Push event, sec (the abnormal network situation)"
PROGRAMING_143 = "Pressing electronic seals"
PROGRAMING_144 = "Cleaning fixation of the effects of magnetic and/or HF fields"
PROGRAMING_145 = "Changing the clock belt"
PROGRAMING_146 = "Changing the sequence of the output to LCD in the 'Auto-Carrot' mode"
PROGRAMING_147 = "Changing the sequence of the output to LCD in the 'button' mode"
PROGRAMING_148 = "Changing the level of the current limit"
PROGRAMING_149 = "Changing the level of the voltage limit"
PROGRAMING_150 = "Emergency tariff number"
PROGRAMING_1024 = "KPZ. Change device parameters"

# for 0.0.96.11.4.255 События внешних воздействий ГОСТ Р 58940—2020 таблица Д.5
EXTERNAL_IMPACT_1 = "Magnetic field - the beginning"
EXTERNAL_IMPACT_2 = "Magnetic field - ending"
EXTERNAL_IMPACT_3 = "Cleansing the electronic seal cover"
EXTERNAL_IMPACT_4 = "Plugging of the electoral seal"
EXTERNAL_IMPACT_5 = "Proceeding the electronic seal of the outer sensor"
EXTERNAL_IMPACT_6 = "The impact of HF fields - beginning"
EXTERNAL_IMPACT_7 = "The impact of HF fields - ending"

# for 0.0.96.11.5.255 Коммуникационные события ГОСТ Р 58940—2020 таблица Д.6
COMMUNICATION_1 = "Connection terminated interface"
COMMUNICATION_2 = "Connection established interface"

# for 0.0.96.11.6.255 События контроля доступа ГОСТ Р 58940—2020 таблица Д.7
ACCESS_1 = "Unauthorized access attempt interface"
ACCESS_2 = "Violation of the protocol requirements"
ACCESS_3 = "Blocking for exceeding the number of incorrect passwords"
ACCESS_4 = "Firmware verification error"

# for 0.0.96.11.7.255 Коды событий для журнала самодиагностики ГОСТ Р 58940—2020 таблица Д.8
SELF_DIAGNOSTIC_1 = "Initializing the Device"
SELF_DIAGNOSTIC_2 = "Measuring unit-error"
SELF_DIAGNOSTIC_3 = "Measuring unit-OK"
SELF_DIAGNOSTIC_4 = "Computing block-error"
SELF_DIAGNOSTIC_5 = "Real-time clock-error"
SELF_DIAGNOSTIC_6 = "Real-time clock-OK"
SELF_DIAGNOSTIC_7 = "Power supply unit-error"
SELF_DIAGNOSTIC_8 = "Power supply unit-OK"
SELF_DIAGNOSTIC_9 = "Display-error"
SELF_DIAGNOSTIC_10 = "Display-OK"
SELF_DIAGNOSTIC_11 = "Memory block-error"
SELF_DIAGNOSTIC_12 = "Memory block-OK"
SELF_DIAGNOSTIC_13 = "Program memory block-error"
SELF_DIAGNOSTIC_14 = "Program memory block-OK"
SELF_DIAGNOSTIC_15 = "Core clocking system-error"
SELF_DIAGNOSTIC_16 = "Core clocking system-OK"
SELF_DIAGNOSTIC_17 = "Clock clocking system-error"
SELF_DIAGNOSTIC_18 = "Clock clocking system-OK"
SELF_DIAGNOSTIC_19 = "Computing unit-OK"

# for 0.0.96.11.8.255 События по превышению реактивной мощности tg ф тангенс сети ГОСТ Р 58940—2020 таблица Д.9
REACTIVE_POWER_1 = "Exceeding the set threshold is the beginning"
REACTIVE_POWER_2 = "Exceeding the set threshold — end"

# for 0.0.96.5.1.255 Статус качества сети профиль суточных показаний
POWER_QUALITY1_1 = "Deviation of voltage by more than 10% of the nominal"
POWER_QUALITY1_2 = "Frequency deviation by more than 0.4 Hz from the nominal"

# for POWER_QUALITY1_0 =.0.96.5.4.255 Статус качества сети журнал качества сети
POWER_QUALITY2_1 = "Reducing the voltage of more than 10%"
POWER_QUALITY2_8 = "Raising voltage for more than 10%"
POWER_QUALITY2_0x10 = "Reducing the frequency of more than 0.4 Hz"
POWER_QUALITY2_0x20 = "Reducing the frequency of more than 0.2 Hz"
POWER_QUALITY2_0x40 = "Increase frequency more than 0.2 Hz"
POWER_QUALITY2_0x80 = "Increase frequency more than 0.4 Hz"
POWER_QUALITY2_0x4000 = "Reducing the frequency of more than a given threshold"
POWER_QUALITY2_0x8000 = "Increase frequency more than a given threshold"



from ..__class_init__ import *
from ...types.implementations.emuns import CommSpeed


class InitializationStringElement(cdt.Structure):
    """ Request - Response strings"""
    values: tuple[cdt.OctetString, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.REQUEST, cdt.OctetString),
                cdt.StructElement(cdt.se.RESPONSE, cdt.OctetString))

    @property
    def request(self) -> cdt.OctetString:
        return self.values[0]

    @property
    def response(self) -> cdt.OctetString:
        return self.values[1]


class InitializationString(cdt.Array):
    """ This data contains all the necessary initialization commands to be sent to the modem in order to  configure it properly. This may include the configuration of special
    modem features.
    If the array contains more than one initialization string element, they are subsequently sent to the modem after receiving an answer matching the defined response.
    It is assumed that the modem is pre-configured so that it accepts the initialization_string . If no initialization is needed, the initialization string is empty."""
    TYPE=InitializationStringElement


class ModemProfileElement(cdt.OctetString):
    """ TODO: can be OK, CONNECT, RING, NO CARRIER, ERROR, CONNECT 1 200, NO DIAL TONE, BUSY, NO ANSWER, CONNECT 600, CONNECT 2 400, CONNECT 4 800, CONNECT 9 600, CONNECT 14 400,
    CONNECT 28 800, CONNECT 36 600, CONNECT 56 000"""

    def __init__(self, value: bytes):
        super(ModemProfileElement, self).__init__(value)
        if self.decode() not in self.get_validate_values:
            raise ValueError(F'Got modem profile element {self.decode()}, expected {b", ".join(self.get_validate_values())}')
        else:
            pass

    @staticmethod
    def get_validate_values() -> tuple[bytes, ...]:
        return b'OK', b'CONNECT', b'RING', b'NO CARRIER', b'ERROR', b'CONNECT 1 200', b'NO DIAL TONE', b'BUSY', b'NO ANSWER', b'CONNECT 600', b'CONNECT 2 400', b'CONNECT 4 800', \
               b'CONNECT 9 600', b'CONNECT 14 400', b'CONNECT 28 800', b'CONNECT 36 600', b'CONNECT 56 000'


class ModemProfile(cdt.Array):
    """ This data defines the mapping from Hayes standard commands/responses to modem specific strings. Shall contain the corresponding stings for the modem used in following 
    order"""
    TYPE = ModemProfileElement


class PSTNModemConfiguration(ic.COSEMInterfaceClasses):
    """ An  instance of the 'PSTN modem configuration' IC stores data related to the initialization of modems, which are used for data transfer from/to a device. Several modems
    can be configured."""
    NAME = cn.PSTN_MODEM_CONFIGURATION
    CLASS_ID = ClassID.MODEM_CONFIGURATION
    VERSION = Version.V0
    A_ELEMENTS = (ic.ICAElement(an.COMM_SPEED, CommSpeed, 0, 9, 5),
                  ic.ICAElement(an.INITIALIZATION_STRING, InitializationString),
                  ic.ICAElement(an.MODEM_PROFILE, ModemProfile))

    def characteristics_init(self):
        """nothing do it"""

    @property
    def comm_speed(self) -> CommSpeed:
        return self.get_attr(2)

    @property
    def initialization_string(self) -> InitializationString:
        return self.get_attr(3)

    @property
    def modem_profile(self) -> ModemProfile:
        return self.get_attr(4)


if __name__ == '__main__':
    b = ModemProfileElement.get_validate_values().index('RING')
    a = ModemProfileElement(b'\x09\x04\x33\x45\x54\x33')



""" COMMON Structs """
from ...types import common_data_types as cdt, cosem_service_types as cst


class ActionItem(cdt.Structure):
    values: tuple[cst.LogicalName, cdt.LongUnsigned]
    ELEMENTS = (cdt.StructElement(cdt.se.SCRIPT_LOGICAL_NAME, cst.LogicalName),
                cdt.StructElement(cdt.se.SCRIPT_SELECTOR, cdt.LongUnsigned))

    @property
    def script_logical_name(self) -> cst.LogicalName:
        return self.values[0]

    @property
    def script_selector(self) -> cdt.LongUnsigned:
        return self.values[1]


class ValueDefinition(cdt.Structure):
    """ Defines an attribute of an object to be monitored. Only attributes with simple data types are allowed. """
    values: tuple[cdt.LongUnsigned, cst.LogicalName, cdt.Integer]
    ELEMENTS = (cdt.StructElement(cdt.se.CLASS_ID, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.LOGICAL_NAME, cst.LogicalName),
                cdt.StructElement(cdt.se.ATTRIBUTE_INDEX, cdt.Integer))

    @property
    def class_id(self) -> cdt.LongUnsigned:
        return self.values[0]

    @property
    def logical_name(self) -> cst.LogicalName:
        return self.values[1]

    @property
    def attribute_index(self) -> cdt.Integer:
        return self.values[2]


class RestrictionByDate(cdt.Structure):
    values: tuple[cst.OctetStringDate, cst.OctetStringDate]
    ELEMENTS = (cdt.StructElement(cdt.se.FROM_DATE, cst.OctetStringDate),
                cdt.StructElement(cdt.se.TO_DATE, cst.OctetStringDate))

    @property
    def from_date(self) -> cst.OctetStringDate:
        return self.values[0]

    @property
    def to_date(self) -> cst.OctetStringDate:
        return self.values[1]


class RestrictionByEntry(cdt.Structure):
    values: tuple[cdt.DoubleLongUnsigned, cdt.DoubleLongUnsigned]
    ELEMENTS = (cdt.StructElement(cdt.se.FROM_ENTRY, cdt.DoubleLongUnsigned),
                cdt.StructElement(cdt.se.TO_ENTRY, cdt.DoubleLongUnsigned))

    @property
    def from_entry(self) -> cdt.DoubleLongUnsigned:
        return self.values[0]

    @property
    def to_entry(self) -> cdt.DoubleLongUnsigned:
        return self.values[1]


class CaptureObjectDefinition(cdt.Structure):
    """ Capture objects that are assigned to the object instance. Upon a call of the capture (data) method or automatically
    in defined intervals, the selected attributes are copied into the buffer of the profile. """
    values: tuple[cdt.LongUnsigned, cst.LogicalName, cdt.Integer, cdt.LongUnsigned]
    default = b'\x02\x04\x12\x00\x08\x09\x06\x00\x00\x01\x00\x00\xff\x0f\x02\x12\x00\x00'
    ELEMENTS = (cdt.StructElement(cdt.se.CLASS_ID, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.LOGICAL_NAME, cst.LogicalName),
                cdt.StructElement(cdt.se.ATTRIBUTE_INDEX, cdt.Integer),
                cdt.StructElement(cdt.se.DATA_INDEX, cdt.LongUnsigned))

    @property
    def class_id(self) -> cdt.LongUnsigned:
        return self.values[0]

    @property
    def logical_name(self) -> cst.LogicalName:
        return self.values[1]

    @property
    def attribute_index(self) -> cdt.Integer:
        """pointer to the attribute within the object. attribute_index 1 refers to the first attribute (i.e. the logical_name), attribute_index 2
        to the 2nd, etc.; attribute_index 0 refers to all public attributes"""
        return self.values[2]

    @property
    def data_index(self) -> cdt.LongUnsigned:
        """pointer selecting a specific element of the attribute. The first element in the attribute structure is identified by data_index 1. If the
        attribute is not a structure, then the data_index has no meaning. If the capture object is the buffer of a profile, then the data_index
        identifies the captured object of the buffer (i.e. the column) of the inner profile. data_index 0: references the whole attribute"""
        return self.values[3]

    def __hash__(self):
        return int.from_bytes(self.logical_name.contents+self.attribute_index.contents+self.data_index.contents, 'big')


class WindowElement(cdt.Structure):
    values: tuple[cst.OctetStringDateTime, cst.OctetStringDateTime]
    ELEMENTS = (cdt.StructElement(cdt.se.START_TIME, cst.OctetStringDateTime),
                cdt.StructElement(cdt.se.END_TIME, cst.OctetStringDateTime))

    @property
    def start_time(self) -> cst.OctetStringDateTime:
        return self.values[0]

    @property
    def end_time(self) -> cst.OctetStringDateTime:
        return self.values[1]



from __future__ import annotations
import dataclasses
from abc import ABC, abstractmethod
from typing import Iterator, Type, TypeAlias, Callable, Any
from .. import settings
from ..types import cdt, ut, cst
from ..relation_to_OBIS import get_name
import logging
from enum import IntEnum
from datetime import datetime, timedelta, timezone
from itertools import count
from . import collection as col

match settings.get_current_language():
    case settings.Language.ENGLISH: from ..Values.EN import attr_names as an
    case settings.Language.RUSSIAN: from ..Values.RU import attr_names as an

logger = logging.getLogger(__name__)
logger.level = logging.INFO

logger.info(F'Register start')

_n_class = count(0)


class Classifier(IntEnum):
    """ (dyn.) Classifies an attribute that carries a process value, which is updated by the meter itself.
    (static) Classifies an attribute, which is not updated by the meter itself (e.g. configuration data). """
    NOT_SPECIFIC = 0
    STATIC = 1
    DYNAMIC = 2


SelectiveAccessDescriptor: TypeAlias = ut.SelectiveAccessDescriptor  # TODO: make with subclass


@dataclasses.dataclass(frozen=True)
class ICElement:
    NAME: str


@dataclasses.dataclass(frozen=True)
class ICAElement(ICElement):
    DATA_TYPE: Type[cdt.CommonDataType] | ut.CHOICE
    min: int = None
    max: int = None
    default: int = None
    classifier: Classifier = Classifier.STATIC
    selective_access: Type[SelectiveAccessDescriptor] | None = None

    def __str__(self):
        return F'{self.NAME} ({self.classifier.name.lower()}) {self.DATA_TYPE.NAME}'


@dataclasses.dataclass(frozen=True)
class ICMElement(ICElement):
    DATA_TYPE: Type[cdt.CommonDataType]

    def __str__(self):
        return F'{self.NAME} {self.DATA_TYPE.NAME}'


_LN_ELEMENT = ICAElement(an.LOGICAL_NAME, cst.LogicalName)
"""" first element for each COSEM Interface Class"""


class COSEMInterfaceClasses(ABC):
    CLASS_ID: ut.CosemClassId
    VERSION: cdt.Unsigned | None = None
    """ Identification code of the version of the class. The version of each object is retrieved together with the logical name and the class_id by reading the object_list 
    attribute of an “Association LN” / ”Association SN” object. Within one logical device, all instances of a certain class must be of the same version."""
    A_ELEMENTS: tuple[ICAElement, ...]
    M_ELEMENTS: tuple[ICMElement, ...] = None
    cardinality: tuple[int, int | None]
    __attributes: list[cdt.CommonDataType | None]
    __specific_methods: tuple[cdt.CommonDataType, ...] = None
    _cbs_attr_post_init: dict[int, Callable]
    __record_time: list[cdt.DateTime | None]  # TODO: make to int
    collection: col.Collection | None

    def __init__(self, logical_name: cst.LogicalName | bytes | str):
        self.collection = None
        # """ TODO: """
        self.cardinality = (0, None)
        """ (min, max). default is (0, None) from 0 to infinity. If min == max then they are value.   
        Specifies the number of instances of the class within a logical device. value The class shall be 
        instantiated exactly “value” times. min...max. The class shall be instantiated at least “min.” times 
        and at most “max.” times. If min. is zero (0) then the class is optional, otherwise (min. > 0) "min." 
        instantiations of the class are mandatory. """

        self.__attributes = [_LN_ELEMENT.DATA_TYPE(logical_name), *[None] * len(self.A_ELEMENTS)]
        """ Attributes container """

        if self.M_ELEMENTS is not None:
            self.__specific_methods = tuple(el.DATA_TYPE() for el in self.M_ELEMENTS)
            """Specific methods container"""

        self._cbs_attr_post_init = dict()
        """container with callbacks for post initial attribute by index"""

        self.__record_time = [None] * len(self.A_ELEMENTS)

        # init all attributes with default value
        for i in range(2, len(self.A_ELEMENTS)+2):
            default = self.get_attr_element(i).default
            if default is not None:
                self.set_attr(i, default)

        self.characteristics_init()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.hash_ = next(_n_class)

    @classmethod
    def get_attr_element(cls, i: int) -> ICAElement:
        """return element by order index. Override in each new class"""
        match i:
            case 1: return _LN_ELEMENT
            case _: return cls.A_ELEMENTS[i - 2]

    @classmethod
    def get_meth_element(cls, i: int) -> ICElement:
        """ implement in subclasses with methods """
        return cls.M_ELEMENTS[i - 1]

    @abstractmethod
    def characteristics_init(self):
        """ initiate all attributes and methods of class """

    def get_attr(self, index: int) -> Any | None:
        if index >= 1:
            return self.__attributes[index-1]
        else:
            raise IndexError(F'not support {index=} as attribute')

    def set_attr(self, index: int, value, with_time: bool | datetime = False) -> timedelta:
        if self.__attributes[index-1] is None:
            self.__attributes[index-1] = self.get_attr_element(index).DATA_TYPE(value if value is not None else self.get_attr_element(index).default)
            match self._cbs_attr_post_init.pop(index, None):
                case None:         """without callback post init"""
                case _ as cb_func: cb_func()
        else:
            self.__attributes[index-1].set(value)
        # todo: make better all below
        TZ = timezone(datetime.now() - datetime.utcnow())
        """ os time zone """
        time_now = datetime.now(TZ)
        if isinstance(with_time, bool) and with_time:
            self.set_record_time(index, cdt.DateTime(time_now))
            return timedelta()
        elif isinstance(with_time, datetime):
            delta = (time_now-with_time)/2
            self.set_record_time(index, cdt.DateTime(time_now - delta))
            return delta

    def set_attr_link(self, index: int, link: cdt.CommonDataType):
        # self.__attributes[index - 1] = link  # TODO: without validate now for pass load_objects
        if isinstance(link, self.get_attr_element(index).DATA_TYPE):
            self.__attributes[index-1] = link
        else:
            raise ValueError(F'get wrong link: {link} for {self} attr: {index}')

    def get_attr_data_type(self, index: int) -> Type[cdt.CommonDataType] | ut.CHOICE:
        """search data_type attribute value"""
        value: cdt.CommonDataType = self.get_attr(index)
        if value is not None:
            return value.__class__
        else:
            return self.get_attr_element(index).DATA_TYPE

    def clear_attr(self, i: int):
        """use in template"""
        if i > 1:
            self.__attributes[i-1] = None
        else:
            raise ValueError(F'not support clear {self} attr: {i}')

    def get_meth(self, index: int) -> Any:
        if index >= 1:
            return self.__specific_methods[index-1]
        else:
            raise IndexError(F'not support {index=} as attribute')

    def get_record_time(self, index: int) -> cdt.DateTime | None:
        return self.__record_time[index-2]

    def set_record_time(self, index: int, value: str | bytes | cdt.DateTime):
        self.__record_time[index-2] = cdt.DateTime(value)

    def get_index_with_attributes(self, in_init_order: bool = False) -> Iterator[tuple[int, cdt.CommonDataType | None]]:
        """ if by initiation order is True then need override method for concrete class"""
        return iter(zip(range(1, self.get_attr_length()+1), self.__attributes))

    def get_attr_length(self) -> int:
        """common attributes amount"""
        return len(self.A_ELEMENTS)+1

    @property
    def it_index_with_meth(self) -> Iterator[tuple[int, cdt.CommonDataType]]:
        return iter(zip(range(1, 20), self.__specific_methods))

    @property
    @abstractmethod
    def NAME(self) -> str:
        """ class name according to current language """

    @property
    def logical_name(self) -> cst.LogicalName:
        """ The logical name is always the first attribute of a class. It identifies the instantiation (COSEM object) of this class.
        The value of the logical_name conforms to OBIS (see IEC 62056-61)"""
        return self.get_attr(1)

    def __setattr__(self, key, value):
        match key:
            case 'VERSION' | 'CLASS_ID' | 'A_ELEMENTS' | 'M_ELEMENTS' as prop: raise ValueError(F"Don't support set {prop}")
            case _:                                                            super().__setattr__(key, value)

    def __getitem__(self, item) -> cdt.CommonDataType:
        """ get attribute value by index, start with 1 """
        if isinstance(item, str):
            return self.__getattr__(item)
        return self.get_attr(item)

    def __iter__(self) -> Iterator[cdt.CommonDataType]:
        """ return attributes iterator"""
        return iter(self.__attributes)

    def __str__(self):
        return F'{self.logical_name} {get_name(self.logical_name)}'

    def get_obis(self) -> bytes:
        """ return obis as bytes[6] """
        return self.logical_name.contents

    @property
    def instance_id(self) -> cdt.OctetString:
        return self.logical_name

    # TODO: rewrite this
    def get_attribute_descriptor(self, index: int) -> bytes:
        """ Cosem-Attribute-Descriptor IS/IEC 62056-53 : 2006, 8.3 Useful types """
        return self.CLASS_ID.contents + self.instance_id.contents + ut.CosemObjectAttributeId(index).contents

    @property
    def string_type_cardinality(self) -> str:
        min_cardinality, max_cardinality = self.cardinality
        if min_cardinality == max_cardinality:
            return str(min_cardinality)
        else:
            max_cardinality = str(max_cardinality) if max_cardinality else 'n'
            return F'{str(min_cardinality)}...{max_cardinality}'

    def reset_attribute(self, index: int):
        """ try set default to value """
        self.set_attr(index, self.get_attr_element(index).default)

    def get_attr_descriptor(self, value: int) -> ut.CosemAttributeDescriptor:
        """ TODO """
        return ut.CosemAttributeDescriptor((self.CLASS_ID,
                                            ut.CosemObjectInstanceId(self.logical_name.contents),
                                            ut.CosemObjectAttributeId(value)))

    def get_meth_descriptor(self, value: str | int) -> ut.CosemMethodDescriptor:
        """ TODO """
        match value:
            case int() as index:
                return ut.CosemMethodDescriptor((ut.CosemClassId(self.CLASS_ID.contents),
                                                 ut.CosemObjectInstanceId(self.logical_name.contents),
                                                 ut.CosemObjectMethodId(index)))

    def __hash__(self):
        return hash(self.logical_name)



from __future__ import annotations
from . import ver0
from ..__class_init__ import *


class SecurityPolicy(cdt.FlagMixin, ver0.SecurityPolicy):
    """ Enforces authentication and/or encryption and/or digital signature using the security algorithms available within security suite. It applies independently for requests and
     responses. When enum value is interpreted as an unsigned, the meaning of each bit is as shown below """
    ELEMENTS = {0: en.UNUSED,
                1: en.UNUSED,
                2: en.AUTHENTICATION_REQUEST,
                3: en.ENCRYPTED_REQUEST,
                4: en.DIGITALLY_SIGNED_REQUEST,
                5: en.AUTHENTICATION_RESPONSE,
                6: en.ENCRYPTED_RESPONSE,
                7: en.DIGITALLY_SIGNED_RESPONSE}


class CertificateEntity(cdt.Enum):
    """TODO:"""
    ELEMENTS = {b'\x00': en.SERVER,
                b'\x01': en.CLIENT_OR_THIRD_PARTY,
                b'\x02': en.CERTIFICATION_AUTHORITY,
                b'\x03': en.OTHER}


class CertificateType(cdt.Enum):
    """TODO:"""
    ELEMENTS = {b'\x00': en.DIGITAL_SIGNATURE,
                b'\x01': en.KEY_AGREEMENT,
                b'\x02': en.TLS,
                b'\x03': en.OTHER}


class SecuritySuite(ver0.SecuritySuite,
                    elements={b'\x01': en.AES_GCM_128_ECDSA_P_256,
                              b'\x02': en.AES_GCM_256_ECDSA_P_384}):
    """Version 0 extension"""


class CertificateInfo(cdt.Structure):
    """ TODO: """
    values: tuple[CertificateEntity, CertificateType, cdt.OctetString, cdt.OctetString, cdt.OctetString, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.CERTIFICATE_ENTITY, CertificateEntity),
                cdt.StructElement(cdt.se.CERTIFICATE_TYPE, CertificateType),
                cdt.StructElement(cdt.se.SERIAL_NUMBER, cdt.OctetString),
                cdt.StructElement(cdt.se.ISSUER, cdt.OctetString),
                cdt.StructElement(cdt.se.SUBJECT, cdt.OctetString),
                cdt.StructElement(cdt.se.SUBJECT_ALT_NAME, cdt.OctetString))

    @property
    def certificate_entity(self) -> CertificateEntity:
        return self.values[0]

    @property
    def certificate_type(self) -> CertificateType:
        return self.values[1]

    @property
    def serial_number(self) -> cdt.OctetString:
        return self.values[2]

    @property
    def issuer(self) -> cdt.OctetString:
        return self.values[3]

    @property
    def subject(self) -> cdt.OctetString:
        return self.values[4]

    @property
    def subject_alt_name(self) -> cdt.OctetString:
        return self.values[5]


class Certificates(cdt.Array):
    """Carries information on the X.509 v3 Certificates available and stored in the server"""
    TYPE = CertificateInfo


class KeyID(ver0.KeyID,
            elements={b'\x03': en.KEK}):
    """Version 0 extension"""


class KeyTransferData(cdt.Structure):
    """ TODO: """
    values: tuple[KeyID, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.KEY_ID, KeyID),
                cdt.StructElement(cdt.se.KEY_WRAPPED, cdt.OctetString))

    @property
    def key_id(self) -> KeyID:
        return self.values[0]

    @property
    def key_wrapped(self) -> cdt.OctetString:
        return self.values[1]


class KeyTransfer(cdt.Array):
    """ Array of key_transfer_data """
    TYPE = KeyTransferData


class KeyAgreementData(cdt.Structure):
    """ TODO: """
    values: tuple[KeyID, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.KEY_ID, KeyID),
                cdt.StructElement(cdt.se.KEY_DATA, cdt.OctetString))

    @property
    def key_id(self) -> KeyID:
        return self.values[0]

    @property
    def key_data(self) -> cdt.OctetString:
        return self.values[1]


class KeyAgreement(cdt.Array):
    """ Array of key_agreement_data """
    TYPE = KeyAgreementData


class KeyPair(cdt.Enum):
    """TODO:"""
    ELEMENTS = {b'\x00': en.DIGITAL_SIGNATURE_KEY_PAIR,
                b'\x01': en.KEY_AGREEMENT_KEY_PAIR,
                b'\x02': en.TLS_KEY_PAIR}


class CertificateIdentificationByEntity(cdt.Structure):
    """ TODO: """
    values: tuple[CertificateEntity, CertificateType, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.CERTIFICATE_ENTITY, CertificateEntity),
                cdt.StructElement(cdt.se.CERTIFICATE_TYPE, CertificateType),
                cdt.StructElement(cdt.se.SYSTEM_TITLE, cdt.OctetString))

    @property
    def certificate_entity(self) -> CertificateEntity:
        return self.values[0]

    @property
    def certificate_type(self) -> CertificateType:
        return self.values[1]

    @property
    def system_title(self) -> cdt.OctetString:
        return self.values[2]


class CertificateIdentificationBySerial(cdt.Structure):
    """ TODO: """
    values: tuple[cdt.OctetString, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.SERIAL_NUMBER, cdt.OctetString),
                cdt.StructElement(cdt.se.ISSUER, cdt.OctetString))

    @property
    def serial_number(self) -> cdt.OctetString:
        return self.values[0]

    @property
    def issuer(self) -> cdt.OctetString:
        return self.values[1]


class CertificateIdentificationType(cdt.Enum):
    """TODO:"""
    ELEMENTS = {b'\x00': en.CERTIFICATE_IDENTIFICATION_ENTITY,
                b'\x01': en.CERTIFICATE_IDENTIFICATION_SERIAL}


class CertificateIdentificationOption(ut.CHOICE):
    TYPE = cdt.Structure
    ELEMENTS = {0: ut.SequenceElement('by entity', CertificateIdentificationByEntity),
                1: ut.SequenceElement('by serial', CertificateIdentificationBySerial)}


certificate_identification_option = CertificateIdentificationOption()


class CertificateIdentification(cdt.Structure):
    """Override several methods of cdt.Structure. It limited Structure."""
    values: tuple[CertificateIdentificationType, CertificateIdentificationByEntity | CertificateIdentificationBySerial]

    def __init__(self):
        self.__dict__["values"] = (CertificateIdentificationType(0), CertificateIdentificationByEntity())
        self.certificate_identification_type.register_cb_post_set(self.__set_certificate_identification_type)

    def __set_certificate_identification_type(self):
        self.__dict__["values"] = (self.certificate_identification_type, certificate_identification_option(int(self.certificate_identification_type)))

    @property
    def ELEMENTS(self) -> tuple[cdt.StructElement, cdt.StructElement]:
        return (cdt.StructElement(cdt.se.CERTIFICATE_IDENTIFICATION_TYPE, CertificateIdentificationType),
                cdt.StructElement(cdt.se.CERTIFICATION_IDENTIFICATION_OPTIONS,
                                  CertificateIdentificationByEntity if int(self.certificate_identification_type) == 0 else CertificateIdentificationBySerial))

    @property
    def certificate_identification_type(self) -> CertificateIdentificationType:
        return self.values[0]

    @property
    def certificate_identification_options(self) -> CertificateIdentificationByEntity | CertificateIdentificationBySerial:
        return self.values[1]


class SecuritySetup(ver0.SecuritySetup):
    VERSION = Version.V1
    A_ELEMENTS = (ic.ICAElement(an.SECURITY_POLICY, SecurityPolicy),
                  ic.ICAElement(an.SECURITY_SUITE, SecuritySuite),
                  ver0.SecuritySetup.get_attr_element(4),
                  ver0.SecuritySetup.get_attr_element(5),
                  ic.ICAElement(an.CERTIFICATES, Certificates))

    M_ELEMENTS = (ic.ICMElement(mn.SECURITY_ACTIVATE, SecurityPolicy),
                  ic.ICMElement(mn.KEY_TRANSFER, KeyTransfer),
                  ic.ICMElement(mn.KEY_AGREEMENT, KeyAgreement),
                  ic.ICMElement(mn.GENERATE_KEY_PAIR, KeyPair),
                  ic.ICMElement(mn.GENERATE_CERTIFICATE_REQUEST, KeyPair),
                  ic.ICMElement(mn.IMPORT_CERTIFICATE, cdt.OctetString),
                  ic.ICMElement(mn.EXPORT_CERTIFICATE, CertificateIdentification),
                  ic.ICMElement(mn.REMOVE_CERTIFICATE, CertificateIdentification))

    def characteristics_init(self):
        self.set_attr(2, None)
        self.set_attr(3, None)
        self.set_attr(6, None)

    @property
    def security_policy(self) -> SecurityPolicy:
        return self.get_attr(2)

    @property
    def security_suite(self) -> SecuritySuite:
        return self.get_attr(3)

    @property
    def certificates(self) -> Certificates:
        return self.get_attr(6)

    @property
    def security_activate(self) -> SecurityPolicy:
        return self.get_meth(1)

    @property
    def key_transfer(self) -> KeyTransferData:
        return self .get_meth(2)

    def key_agreement(self) -> KeyAgreement:
        return self .get_meth(3)

    def generate_certificate_request(self) -> KeyPair:
        return self .get_meth(4)

    def import_certificate(self) -> cdt.OctetString:
        return self .get_meth(5)

    def export_certificate(self) -> CertificateIdentification:
        return self .get_meth(6)

    def remove_certificate(self) -> CertificateIdentification:
        return self .get_meth(7)
